
/*===========================================================================
 * Lapis language, a object-oriented stack-based language interpreter.
 * Version 0.7
 * Copyright (C) 2005,2011 by Zachary Smith.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 1, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *=========================================================================*/

/*---------------------------------------------------------------------------
 * Change log
 *
 * 2005,2011 Aug 10: added string manipulation functionality.
 * 2005,2011 Aug 10: added stack manipulation functionality.
 * 2005,2011 Aug 10: added styles to table elements.
 * 2005,2011 Aug 15: added image creation, printing and basic drawing operations.
	=> 0.5
 * 2005,2011 Aug 16: added advanced math functions.
 * 2005,2011 Aug 16: added matrix & dictionary creation & printing.
 * 2005,2011 Aug 16: dictionary is now sorted.
 * 2005,2011 Aug 16: added html image print.
 * 2005,2011 Aug 16: added boolean type.
 * 2005,2011 Aug 16: added getenv.
	=> 0.6
 * 2005,2011 Aug 20: added array.
 * 2005,2011 Aug 20: improved command execution speed with hash.
 * 2005,2011 Aug 20: added check for reserved word used as variable.
 * 2005,2011 Aug 20: improved speed by combining like cmd operations.
	=> 0.7
 * 2005,2011 Oct 20: finalized (preset) object-oriented features.
 *--------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <math.h>
#include <signal.h>

#include "defs.h"
#include "lapis.h"


#define MEMORY_LOG_SIZE 60000
static unsigned long memory_log [MEMORY_LOG_SIZE];
static unsigned char memory_flags [MEMORY_LOG_SIZE];
static unsigned long memory_ix = 0;


unsigned char *mymalloc (int size)
{
	if (size <= 0)
		FATAL("zero/negative malloc size");

	unsigned char *p = malloc(size);
	if (!p)
		FATAL("malloc failed");

	if (memory_ix >= MEMORY_LOG_SIZE)
		FATAL("memory log full");
	memory_flags [memory_ix] = 0;
	memory_log [memory_ix] = (unsigned long) p;
//printf ("mymalloc: stored $%08lx in spot %d\n", (unsigned long) p, (int)memory_ix);
	memory_ix++;

	return p;
}

char *mystrdup (char *s)
{
	int len = strlen(s);
	char *s2 = mymalloc (len+2);
	strcpy (s2,s);
	return s2;
}

void myfree (void *ptr)
{
	if (!ptr)
		FATAL("attempt to free null ptr");

	unsigned long p = (unsigned long) ptr;
//printf ("myfree: ptr=$%08lx\n", p);
	int i=0;
	while (i < memory_ix) {
		if (p == memory_log[i]) {
			if (memory_flags[i] & 1) {
				WARNING("attempt to re-free ptr");
				int *ii=0; *ii=0;
				return;
			} else {
				memory_flags[i] = 1;
				//free (ptr);
				return;
			}
		}
		i++;
	}
	WARNING("attempt to free ptr that was never allocated");
	int *ii=0; *ii=0;
}


#include <jpeglib.h>
JSAMPLE * jpeg_image_data; /* Points to large array of R,G,B-order data */
int jpeg_image_height;     /* Number of rows in image */
int jpeg_image_width;      /* Number of columns in image */
extern void write_JPEG_file (char * filename, int quality);


#ifdef SUPPORT_LAPIS


static Dictionary *global_vars = NULL;
static char content = 0;
static FILE *hef = NULL;
static char *outpath2 = NULL;

// static int class_counter = 0;

static Class *classes[27] = {
	NULL, NULL, NULL, NULL, NULL,
	NULL, NULL, NULL, NULL, NULL,
	NULL, NULL, NULL, NULL, NULL,
	NULL, NULL, NULL, NULL, NULL,
	NULL, NULL, NULL, NULL, NULL,
	NULL, NULL, 
};


bool process_lapis_core (int, char *, char *, char *, char*);
void StackItem_delete (StackItem *im);
void StackItem_print (StackItem*,FILE*f);


Font* Font_load_consolefont (char *name)
{
	if (!name)
		return NULL;

	char path[MAXPATH];
	sprintf (path, "/usr/share/kbd/consolefonts/%s.fnt", name);

	int size = stat_file (path);
	if (size != 4096)	// 8x16 font 
		return NULL;

	FILE *f = fopen (path, "rb");
	if (!f)
		return NULL;

	Font *font = (Font*) mymalloc(sizeof(Font));
	if (!font)
		return NULL;

	uchar *buf = mymalloc(4096);
	if (!buf) {
		myfree (font);
		return NULL;
	}

	if (4096 != fread (buf, 1, 4096, f)) {
		myfree (buf);
		return NULL;
	}

	font->data = buf;
	font->width = 8;
	font->height = 16;
	font->consolefont = 1;
	font->bold = 0;
	font->italic = 0;
	font->fixed = 1;
	font->refcount = 0;
	font->name = mystrdup(name);
	return font;
}

void Font_delete (Font* font)
{
	if (font) {
		font->refcount--;
		if (font->refcount > 0)
			return;

		if (font->name)
			myfree (font->name);
		if (font->data)
			myfree (font->data);
		myfree (font);
	}
}

Class *Class_new (char *name)
{
	if (!name)
		return NULL;

	Class *class = (Class*) mymalloc (sizeof(Class));
	if (!class)
		return NULL;

	memset (class, 0, sizeof(Class));
	class->name = mystrdup (name);
	return class;
}

Method *Class_get_method (Class *class, char *name)
{
	int ix;
	char ch;
	Method *m;

	if (!name)
		return NULL;
	ch = *name;
	ix = isalpha(ch) ? tolower(ch)-'a' : 26;
	m = class->methods[ix];
	while (m) {
		if (!strcmp (m->name, name)) 
			return m;
		m = m->next;
	}
	return NULL;
}

	
Method *Method_new (char *name, method_func func, int nparams)
{
	if (!name)
		return NULL;

	Method *method = (Method*) mymalloc (sizeof(Method));
	if (!method)
		return NULL;

	memset (method, 0, sizeof(Method));
	method->name = mystrdup (name);
	method->nparams = nparams;
	method->func = func; // can be null
	return method;
}

Method *Method_new2 (char *name, method_func func, int nparams, int cmdnum)
{
	Method *nu = Method_new (name,func,nparams);
	nu->cmdnum = cmdnum;
	return nu;
}

Array *Array_new (unsigned int size)
{
	void **p;
	unsigned int alloc_size = size;

	alloc_size = size & ~15;
	alloc_size += 16;

	Array *array = (Array*) mymalloc (sizeof(Array));
	if (!array)
		return NULL;
	array->size = size;
	array->allocated_size = alloc_size;
	p = array->values = (void**) mymalloc (sizeof(void*) * alloc_size);
	if (!p) {
		myfree(array);
		return NULL;
	}
	memset (p, 0, alloc_size); // clear to white
	return array;
}

void Array_delete (Array *ary)
{
	if (ary) {
		ary->refcount--;
		if (ary->refcount > 0)
			return;

		if (ary->values)
			myfree (ary->values);
		myfree (ary);
	}
}

void Array_print (Array *ary, FILE *f)
{
	if (!ary || !f)
		return;

	int i;
	fprintf (f, "\n\n<!--- array --->\n");
	fprintf (f, "<table border=1>\n");
	for (i=0; i < ary->size; i++) {
		StackItem *im = (StackItem*) ary->values[i];

		fprintf (f, "\t<tr><td>");

		if (!im)
			fprintf (f, "(null)");
		else
			StackItem_print (im, f);
		fprintf (f, "</td></tr>\n");
	}
	fprintf (f, "</table>\n");
}

void Array_put (Array *ary, int index, StackItem *datum)
{
	if (!ary || !datum || index < 0 || index >= ary->size) {
		if (datum)
			StackItem_delete (datum);
		return;
	}

	StackItem *im = ary->values[index];
	if (im)
		StackItem_delete (im);

	ary->values [index] = datum;
}

Matrix *Matrix_new (short w, short h)
{
	long double *p;
	unsigned long size;

	if (w <= 0 || h <= 0)
		return NULL;

	Matrix *matrix = (Matrix*) mymalloc (sizeof(Matrix));
	if (!matrix)
		return NULL;
	matrix->width = w;
	matrix->height = h;
	size = w*h*sizeof(long double);
	p = matrix->values = (long double*) mymalloc (size);
	if (!p) {
		myfree(matrix);
		return NULL;
	}
	memset (p, 0, size); // clear to white
	return matrix;
}

void Matrix_delete (Matrix *matrix)
{
	if (matrix) {
		matrix->refcount--;
		if (matrix->refcount > 0)
			return;

		if (matrix->values)
			myfree(matrix->values);
		myfree (matrix);
	}
}

// This function is provided more for testing than for serious use.
void Matrix_print (Matrix *matrix, FILE *f)
{
	if (!matrix || !f)
		return;

	int i,j;
	int w = matrix->width;
	int h = matrix->height;
	fprintf (f, "\n<!--- matrix --->\n");
	fprintf (f, "<table border=1>\n");
	for (j=0; j<h; j++) {
		fprintf (f, "\t<tr>\n");
		for (i=0; i<w; i++) {
			long double v = matrix->values[i+j*w];
			fprintf (f, "\t\t<td>%Lf\n", v);
		}
	}
	fprintf (f, "</table>\n");
}

void Matrix_put (Matrix *matrix, int row, int column, long double value)
{
	if (!matrix || row < 0 || row >= matrix->height ||
			column < 0 || column >= matrix->width) {
		return;
	}

	matrix->values [column+row*matrix->width] = value;
}

Dictionary *Dictionary_new ()
{
	Dictionary *dictionary = (Dictionary*) mymalloc (sizeof(Dictionary));
	if (!dictionary)
		return NULL;
	memset (dictionary, 0, sizeof(Dictionary));
	return dictionary;
}

void Dictionary_delete (Dictionary *dictionary)
{
	if (dictionary) {
		dictionary->refcount--;
		if (dictionary->refcount > 0)
			return;

// XX need to myfree entries
		myfree (dictionary);
	}
}

void Dictionary_print (Dictionary *dictionary, FILE *f)
{
	if (!dictionary || !f)
		return;

	DictionaryItem *im = dictionary->list;
	fprintf (f, "\n<!--- dictionary --->\n");
	fprintf (f, "<table border=1>\n");

	while (im) {
		fprintf (f, "<tr><td>%s</td><td>\n\t", im->name);
		StackItem_print (im->value, f);
		fprintf (f, "</td></tr>\n");
		im = im->next;
	}
	fprintf (f, "</table>\n");
}

Image *Image_new (short w, short h)
{
	uchar *p;
	unsigned long size;

	if (w <= 0 || h <= 0)
		return NULL;

	Image *img = (Image*) mymalloc (sizeof(Image));
	if (!img)
		return NULL;
	img->width = w;
	img->height = h;
	size = w*h*3;
	p = img->pixels = (uchar*) mymalloc (size);
	if (!p) {
		myfree(img);
		return NULL;
	}
	memset (p, 0xff, size); // clear to white
//printf ("image=$%08lx\n", (unsigned long)img);
	return img;
}

void Image_delete (Image *img)
{
	if (img) {
		img->refcount--;
		if (img->refcount > 0)
			return;

		if (img->pixels)
			myfree(img->pixels);
		myfree (img);
	}
}


void Image_draw_point (Image *img, short x, short y)
{
	if (!img || !img->pixels)
		return;
	if (x < 0 || y < 0 || x >= img->width || y >= img->height)
		return;

	unsigned long rgb = img->color;
	unsigned long ix = y * img->width + x;
	uchar *p = img->pixels;
	ix *= 3;
	p[ix++] = (rgb>>16) & 0xff; // jpeg uses big-endian order
	p[ix++] = (rgb>>8) & 0xff;
	p[ix++] = rgb & 0xff;
}

unsigned long Image_read_point (Image *img, short x, short y)
{
	if (!img || !img->pixels)
		return 0;
	if (x < 0 || y < 0 || x >= img->width || y >= img->height)
		return 0;

	unsigned long ix = y * img->width + x;
	uchar *p = img->pixels;
	ix *= 3;
	unsigned long rgb = p[ix++];
	rgb <<= 8;
	rgb |= p[ix++];
	rgb <<= 8;
	rgb |= p[ix++];
	return rgb;
}

// The in-line image printing routine. Not to be used for complicated images.
//
void Image_print (Image *img, FILE *f)
{
	if (!img || !f) 
		return;

	int x, y;
	int w = img->width;
	int h = img->height;

	fprintf (f, "<!--- image --->\n");
	fprintf (f, "<table cellpadding=0 cellspacing=0>\n");

	for (y=0; y<h; y++) {
		fprintf (f, "<tr>");
		for (x=0; x<w; ) {
			unsigned long rgb = Image_read_point (img,x,y);

			int count=1;
			while (count+x < w) {
				unsigned long rgb2 = Image_read_point 
					(img,x+count,y);
				if (rgb2 != rgb)
					break;
				count++;
			}

			fprintf (f, "<td height=1 width=%d colspan=%d bgcolor=#%06lx><font size=-6> ", count,count,rgb);
			x+=count;
		}
		fprintf (f, "</tr>\n");
	}
	fprintf (f, "</table>\n");
}


void Image_tinyblit (Image *img, unsigned long datum)
{
	if (!img || !img->pixels)
		return;

	short x = img->x;
	short y = img->y;
	while (datum) {
		if (datum & 1)
			Image_draw_point (img, x,y);
		x++;
		datum >>= 1;
	}
}

void Image_draw_char (Image *img, uchar ch)
{
	Font *font;

	if (!img || !img->pixels || !img->font || !img->font->data)
		return;
	font = img->font;
	if (!font->fixed || !font->consolefont || font->width != 8 ||
	    font->height != 16)
		return;

	unsigned long datum;
	unsigned long ix = ch * 16;
	unsigned short i=0;
	while (i < 16) {
		datum = font->data[ix++];
		if (datum) {
			if (font->consolefont) {
				unsigned long d2=0;
				if (datum & 1) d2 |= 128;
				if (datum & 2) d2 |= 64;
				if (datum & 4) d2 |= 32;
				if (datum & 8) d2 |= 16;
				if (datum & 16) d2 |= 8;
				if (datum & 32) d2 |= 4;
				if (datum & 64) d2 |= 2;
				if (datum & 128) d2 |= 1;
				datum = d2;
			}

			Image_tinyblit (img, datum);
		}
		img->y++;
		i++;
	}
	img->y -= 16;
}

void Image_draw_string (Image *img, uchar *str)
{
	if (!img || !str || !img->font)
		return;

	int maxcount = 1000;
	int orig_x = img->x;
	while (*str && maxcount>0) {
		Image_draw_char (img, *str++);
		img->x += img->font->width;
		maxcount--;
	}
	img->x = orig_x;
}

void Image_lineto (Image *img, short x1, short y1)
{
	int x, y, s1, s2, dx, dy, xchange, e, j;
	if (!img)
		return;

        s1 = 1;
        s2 = 1;
        x = img->x;
        y = img->y;

        dx = x1-x;
        if (dx < 0) {
                dx = -dx;
                s1 = -1;
        }

        dy = y1-y;
        if (dy < 0) {
                dy = -dy;
                s2 = -1;
        }

        xchange = 0;

        if (dy>dx) {
                int tmp = dx;
                dx = dy;
                dy = tmp;
                xchange = 1;
        }

        e = (dy<<1) - dx;
        for (j=0; j <= dx; j++)
        {
                Image_draw_point (img,x,y);

                if (e >= 0) {
                        if (xchange)
                                x += s1;
                        else
                                y += s2;
                        e -= (dx << 1);
                }
                if (xchange)
                        y += s2;
                else
                        x += s1;
                e += (dy << 1);
        }

	img->x = x1;
	img->y = y1;
}

void Image_fillrect (Image *img, short x1, short y1)
{
	short x,y;
	if (!img)
		return;

	if (img->x > x1 || img->y > y1)
		return;
	for (x=img->x; x<=x1; x++)
		for (y=img->y; y<=y1; y++)
			Image_draw_point (img,x,y);

	img->x = x1;
	img->y = y1;
}


DictionaryItem *DictionaryItem_new ()
{
	DictionaryItem *nu = (DictionaryItem*) mymalloc (sizeof(DictionaryItem));
	if (!nu) 
		return NULL;
	memset (nu, 0, sizeof (DictionaryItem));
	return nu;
}

void DictionaryItem_delete (DictionaryItem *im)
{
	if (!im) return;

	if (im->value)
		StackItem_delete (im->value);
	if (im->name)
		myfree (im->name);
	myfree (im);
}


int Dictionary_remove (Dictionary *dictionary, char *name)
{
	DictionaryItem *im;
	DictionaryItem *prev=NULL;

	if (!dictionary || !name)
		return 0;

	im = dictionary->list;
	if (!im)
		return 0;

	while (im && strcmp (name, im->name)) {
		prev = im;
		im = im->next;
	}

	if (im) {
		if (!prev)
			dictionary->list = im->next;
		else
			prev->next = im->next;

		im->next = NULL;
		DictionaryItem_delete (im);
		return 1;
	} else
		return 0;
}


StackItem *Dictionary_get (Dictionary *dictionary, char *name)
{
	DictionaryItem *im;

	if (!dictionary || !name)
		return NULL;
//printf ("Dictionary_get for %s\n", name);

	im = dictionary->list;

	while (im && strcmp (name, im->name)) {
		im = im->next;
	}
	return im ? im->value : NULL;
}

bool Dictionary_put (Dictionary *dictionary, char *name, StackItem *value)
{
	if (!dictionary || !name || !value)
		return false;

	if (value->type == STACKITEM_DICTIONARY && 
	    dictionary == value->datum.dictionary) {
		warning ("attempt to store dictionary inside itself");
		return false;
	}

	Dictionary_remove (dictionary, name);
	
	DictionaryItem *im = DictionaryItem_new ();
	im->name = mystrdup(name);
	im->next = NULL;
	im->value = value;

	DictionaryItem *i = dictionary->list;
	DictionaryItem *prev = NULL;

	if (!i) {
		dictionary->list = im;
		return true;
	}

	while (i && strcmp(name,i->name) > 0) {
		prev = i;
		i = i->next;
	}

	if (prev) 
		prev->next = im;
	else 
		dictionary->list = im;

	im->next = i;
	return true;
}


int getword (FILE *f, char *buf, int max)
{
	if (!f || !buf)
		return -1;
	/*---------*/

	int ix=0;
	int ch = fgetc(f);
	while (ch == ' ' || ch == '\t' || ch == '\r' || ch == '\n')
		ch = fgetc(f);

	bool in_quote = ch == '"';
	if (in_quote) {
		buf[ix++] = ch;
		ch = fgetc(f);
	}

	if (ch == EOF)
		return -1;
	if (ch == '"') {
		buf[ix] = 0;
		return 1;
	}

	buf[ix++] = ch;
	ch = fgetc(f);
	while (in_quote||(ch != ' ' && ch != '\t' && ch != '\r' && ch != '\n'))
	{
		if (ch == EOF)
			break;
		else
		if (in_quote ) {
			if (ch == '\\')
				ch = fgetc(f);
			else
			if (ch == '"')
				break;
		}

		if (ix < (max-1))
			buf[ix++] = ch;
		ch = fgetc(f);
	}
	buf[ix] = 0;

	if (!ix && ch==EOF)
		return -1;
	return ix;
}

StackItem *StackItem_new () 
{
	StackItem *nu = (StackItem*) mymalloc (sizeof (StackItem));
//printf ("StackItem_new: nu=$%08lx\n", (unsigned long)nu);
	if (!nu)
		return NULL;
	memset (nu, 0, sizeof (StackItem));
	return nu;
}

void StackItem_delete (StackItem *im)
{
	if (im) {
//printf ("StackItem_delete: im=$%08lx, type=%d\n", (unsigned long)im, im->type);
		switch (im->type) {
		case STACKITEM_STRING:
		case STACKITEM_STYLE:
		case STACKITEM_SYMBOL:
			/* can be null in case of nostyle */
			if (im->datum.string)
				myfree (im->datum.string);
			break;
		
		case STACKITEM_IMAGE:
			Image_delete (im->datum.image);
			break;
		case STACKITEM_MATRIX:
			Matrix_delete (im->datum.matrix);
			break;
		case STACKITEM_FONT:
			Font_delete (im->datum.font);
			break;
		case STACKITEM_ARRAY:
			Array_delete (im->datum.array);
			break;
		case STACKITEM_DICTIONARY:
			Dictionary_delete (im->datum.dictionary);
			break;
		}

		myfree (im);
	}
}

StackItem *StackItem_dup (StackItem *im)
{
	if (!im)
		return NULL;

	StackItem *nu = StackItem_new();
	if (!nu)
		exit(100);

	nu->type = im->type;
	switch (im->type) {
	case STACKITEM_INTEGER:
		nu->datum.integer = im->datum.integer;
		break;
	case STACKITEM_REAL:
		nu->datum.real = im->datum.real;
		break;
	case STACKITEM_STRING:
	case STACKITEM_STYLE:
	case STACKITEM_SYMBOL:
		if (im->datum.string)
			nu->datum.string = (uchar*) mystrdup (im->datum.string);
		else
			; // XX Report this error 
		break;
	case STACKITEM_ARRAY:
		nu->datum.array = im->datum.array;
		nu->datum.array->refcount++;
		break;
	case STACKITEM_DICTIONARY:
		nu->datum.dictionary = im->datum.dictionary;
		nu->datum.dictionary->refcount++;
		break;
	case STACKITEM_IMAGE:
		nu->datum.image = im->datum.image;
		nu->datum.image->refcount++;
		break;
	case STACKITEM_MATRIX:
		nu->datum.matrix = im->datum.matrix;
		nu->datum.matrix->refcount++;
		break;
	case STACKITEM_FONT:
		nu->datum.font = im->datum.font;
		nu->datum.font->refcount++;
		break;
	}

	return nu;
}

void StackItem_tostring (StackItem *im)
{
	if (!im)
		return;
	if (im->type == STACKITEM_STRING)
		return;
	if (im->type == STACKITEM_SYMBOL || im->type == STACKITEM_STYLE) {
		im->type = STACKITEM_STRING;
		return;
	}

	char str[64];
	str[0]=0;
	switch (im->type) {
	case STACKITEM_INTEGER:
		sprintf (str, "%d", im->datum.integer);
		break;
	case STACKITEM_REAL:
		sprintf (str, "%Lf", im->datum.real);
		break;
	}

	im->type = STACKITEM_STRING;
	im->datum.string = (uchar*) mystrdup (str);
}

void StackItem_print (StackItem *im, FILE *ouf)
{
	if (!im || !ouf)
		return;

	if (im->type == STACKITEM_NULL) {
		fprintf (ouf, "null");
		return;
	}

	if (im->title)
		fprintf (ouf, "<title>");
	if (im->bold)
		fprintf (ouf, "<b>");
	if (im->italic)
		fprintf (ouf, "<i>");
	if (im->underline)
		fprintf (ouf, "<u>");
	if (im->strikeout)
		fprintf (ouf, "<s>");
	if (im->super)
		fprintf (ouf, "<sup>");
	if (im->sub)
		fprintf (ouf, "<sub>");
	if (im->size)
		fprintf (ouf, "<font size=%s%d>", im->size>1 ? "+" : "", 
			im->size);

	switch (im->type) {
	case STACKITEM_INTEGER:
		fprintf (ouf, "%d", im->datum.integer);
		break;
	case STACKITEM_REAL:
		fprintf (ouf, "%Lf", im->datum.real);
		break;
	case STACKITEM_STRING:
		fprintf (ouf, "%s", im->datum.string);
		break;
	case STACKITEM_SYMBOL:
		fprintf (ouf, "/%s", im->datum.string);
		break;
	case STACKITEM_IMAGE:
		Image_print (im->datum.image, ouf);
		break;
	case STACKITEM_MATRIX:
		Matrix_print (im->datum.matrix, ouf);
		break;
	case STACKITEM_ARRAY:
		Array_print (im->datum.array, ouf);
		break;
	case STACKITEM_NULL:
		fprintf (ouf, "null");
		break;
	case STACKITEM_ARY_BOUNDARY:
		fprintf (ouf, "(array boundary)");
		break;
	case STACKITEM_DICTIONARY:
		Dictionary_print (im->datum.dictionary, ouf);
		break;
	case STACKITEM_BOOLEAN:
		fprintf (ouf, "%s", im->datum.integer ? "true" : "false");
		break;
	}

	if (im->size)
		fprintf (ouf, "<font size=%s%d>", im->size>1 ? "" : "+",
			-im->size);
	if (im->sub)
		fprintf (ouf, "</sub>");
	if (im->super)
		fprintf (ouf, "</sup>");
	if (im->strikeout)
		fprintf (ouf, "</s>");
	if (im->underline)
		fprintf (ouf, "</u>");
	if (im->italic)
		fprintf (ouf, "</i>");
	if (im->bold)
		fprintf (ouf, "</b>");
	if (im->title)
		fprintf (ouf, "</title>\n");
}

Table* Table_new () {
	Table *t = (Table*) mymalloc(sizeof(Table));
	bzero ((void*)t, sizeof (Table));

	int i;
	for (i=0; i<MAX_COLUMNS; i++)
		t->column_is_numeric[i] = true; // flags to be cleared during parsing

	return t;
}

TableRow* TableRow_new () {
	TableRow *tr = (TableRow*) mymalloc(sizeof(TableRow));
	bzero ((void*)tr, sizeof (TableRow));
	return tr;
}

TableItem* TableItem_new () {
	TableItem *ti = (TableItem*) mymalloc(sizeof(TableItem));
	bzero ((void*)ti, sizeof (TableItem));
	return ti;
}

bool Table_add_header (Table *t, TableRow *r) {
	if (!t || !r)
		return false;
	/*----------*/

	if (!t->header) {
		t->header = r;
		return true;
	} else
		return false;
}

void Table_add_row (Table *t, TableRow *r) {
	if (!t || !r)
		return;
	/*----------*/

	int i=0;
	TableItem *im = r->items;
	while (im && i < MAX_COLUMNS) {
		if (im->type != 1)
			t->column_is_numeric[i] = false;
		i++;
		im =im->next;
	}

	if (!t->rows) {
		t->rows = r;
	} else {
		t->last->next = r;
	}
	t->last = r;
	++t->total_rows;
	r->next = NULL;
}

void TableRow_add_item (Table *t, TableRow *r, TableItem *item) {
	if (!t || !r || !item)
		return;
	/*----------*/

	if (!r->items) {
		r->items = item;
	} else {
		r->last->next = item;
	}
	r->last = item;
	++r->total_items;
	if (r->total_items > t->total_columns)
		t->total_columns = r->total_items;
	item->next = NULL;
}


int cmd_fatal (short cmdnum, StackItem **stackp, FILE *ouf)
{
	StackItem *tos = *stackp;

	if (!tos)
		return LAPIS_ERR_UNDERFLOW;
	if (tos->type != STACKITEM_STRING)
		return LAPIS_ERR_TYPE;

	FATAL(tos->datum.string)

	return LAPIS_ERR_FATAL;
}

int cmd_exit (short cmdnum, StackItem **stackp, FILE *ouf)
{
	StackItem *tos = *stackp;

	if (!tos)
		return LAPIS_ERR_UNDERFLOW;
	if (tos->type != STACKITEM_INTEGER)
		return LAPIS_ERR_TYPE;
	exit (tos->datum.integer);

	return 0;
}

int cmd_lbrace (short cmdnum, StackItem **stackp, FILE *ouf)
{
	return 0;
}

int cmd_rbrace (short cmdnum, StackItem **stackp, FILE *ouf)
{
	return 0;
}

int method_Image_clear (short cmdnum, StackItem *self, StackItem *in, StackItem **out, FILE *ouf)
{
	Image *img = self->datum.image;
	memset (img->pixels, 0xff, img->width * img->height * 3);
	return 0;
}

int method_print (short cmdnum, StackItem *self, StackItem *in, StackItem **out, FILE *ouf)
{
	StackItem_print (self, ouf);
	return LAPIS_OK_REMOVE_SELF;
}

int method_sqrt (short cmdnum, StackItem *self, StackItem *in, StackItem **out, FILE *ouf)
{
	long double value;
	if (self->type == STACKITEM_INTEGER)
		value = (long double) self->datum.integer;
	else
	if (self->type == STACKITEM_REAL)
		value = self->datum.real;
	else
		return LAPIS_ERR_TYPE;

	value = sqrt (value);

	if (self->type == STACKITEM_INTEGER)
		self->datum.integer = (int) value;
	else
	if (self->type == STACKITEM_REAL)
		self->datum.real = value;

	return LAPIS_SUCCESS;
}

int method_write (short cmdnum, StackItem *self, StackItem *in, StackItem **out, FILE *ouf)
{
	StackItem_print (self, ouf);
	return LAPIS_SUCCESS;
}



int method_Font_new (short cmdnum, StackItem *self, StackItem *in, StackItem **out, FILE *ouf)
{
	StackItem *nu = StackItem_new();
	nu->type = STACKITEM_FONT;
	nu->datum.font = NULL;
	nu->next = NULL;
	*out = nu;
	return 0;
}
int method_String_new (short cmdnum, StackItem *self, StackItem *in, StackItem **out, FILE *ouf)
{
	StackItem *nu = StackItem_new();
	nu->type = STACKITEM_STRING;
	nu->datum.string = mystrdup ("");
	nu->next = NULL;
	*out = nu;
	return 0;
}

int method_String_reverse (short cmdnum, StackItem *self, StackItem *in, StackItem **out, FILE *ouf)
{
	StackItem *nu = StackItem_new ();

	char *s = in->datum.string;
	int len = strlen (s);

	nu->datum.string = s;
	in->datum.string = NULL;

	*out = nu;

	int i=0;
	int j=len-1;
	while (i < j) {
		char tmp=s[i];
		s[i]=s[j];
		s[j]=tmp;
		i++;
		j--;
	}

	return 0;
}

// Multiple string ops
//
int method_String (short cmdnum, StackItem *self, StackItem *in, StackItem **out, FILE *ouf)
{
	StackItem *tos = in;

	if (tos->type != STACKITEM_INTEGER &&
	    tos->type != STACKITEM_REAL) {
//printf ("tos->type=%d\n", tos->type);
		return LAPIS_ERR_TYPE;
	}
	if (self->type != STACKITEM_STRING)  {
//printf ("self->type=%d\n", self->type);
		return LAPIS_ERR_TYPE;
	}

	int index = tos->type == STACKITEM_INTEGER ? tos->datum.integer 
				: (int) tos->datum.real;

	char *str = self->datum.string;
	self->datum.string = NULL;
	int len = strlen (str);

	StackItem *nu = StackItem_new ();
	nu->type = STACKITEM_STRING;
	nu->datum.string = str;

	switch (cmdnum) {
	case CMD_STRINGCHAR:
		nu->type = STACKITEM_INTEGER;
		if (index < 0 || index >= len)
			nu->datum.integer = 0;
		else
			nu->datum.integer = nu->datum.string[index];
		myfree (str);
		break;

	case CMD_STRINGLEFT:
		if (index < 0)
			index = 0;
		if (index < len) {
			str[index] = 0;
		}
		break;

	case CMD_STRINGRIGHT:

		if (index < 0)
			index = 0;
		if (!index) 
			str[0] = 0;
		else
		if (index < len) {
			int ix = len - index;
			int ix2 = 0;	
			while ((str[ix2++] = str[ix++]))
				;
			str[ix2] = 0;
		}
	}

	*out = nu;

	return 0;
}


int cmd_style (short cmdnum, StackItem **stackp, FILE *ouf)
{
	StackItem *tos = *stackp;
	tos->type = STACKITEM_STYLE;
	return 0;
}


int method_String_sub (short cmdnum, StackItem *self, StackItem *in, StackItem **out, FILE *ouf)
{
	StackItem *tos = in;
	StackItem *nos = tos->next;
	StackItem *third = nos->next;
	int len, param1, param2;
	bool scrub=false;

	if (tos->type != STACKITEM_INTEGER || nos->type != STACKITEM_INTEGER
	 || third->type != STACKITEM_STRING)
		return LAPIS_ERR_TYPE;

	char *str = third->datum.string;
	third->datum.string = NULL;
	len = strlen (str);

	param1 = nos->datum.integer;
	if (param1 < 0)
		param1 = 0;
	if (param1 >= len) 
		scrub = true;

	if (!scrub) {
		param2 = tos->datum.integer;
		if (param2 < 0) 
			scrub = true;
		if (param2 >= len)
			param2 = len-1;
	}

	if (scrub || param1 > param2)
		str[0] = 0;
	else {
		int ix = len - param1;
		int ix2 = 0;	
		char ch;
		while (ix <= param2 && (ch = str[ix++]))
			str[ix2++] = ch;
		str[ix2] = 0;
	}

	StackItem *nu = StackItem_new();
	nu->type = STACKITEM_STRING;
	nu->datum.string = str;

	*out = nu;

	return 0;
}


int method_Image_show (short cmdnum, StackItem *self, StackItem *tos, StackItem **out, FILE *ouf)
{
	if (tos->type != STACKITEM_STRING ||
	    self->type != STACKITEM_IMAGE)
		return LAPIS_ERR_TYPE;

	Image_draw_string (self->datum.image,
		tos->datum.string);

	return 0;
}

int method_Image_setfont (short cmdnum, StackItem *self, StackItem *tos, StackItem **out, FILE *ouf)
{
	Font *font;

	if (tos->type != STACKITEM_FONT || 
	    self->type != STACKITEM_IMAGE)
		return LAPIS_ERR_TYPE;

	font = self->datum.image->font = tos->datum.font;
	font->refcount++;

	return 0;
}


int method_Image_setcolor (short cmdnum, StackItem *self, StackItem *tos, StackItem **out, FILE *ouf)
{
	if (tos->type != STACKITEM_INTEGER ||
	    self->type != STACKITEM_IMAGE) 
		return LAPIS_ERR_TYPE;

	self->datum.image->color = tos->datum.integer;

	return 0;
}

int cmd_closearray (short cmdnum, StackItem **stackp, FILE *ouf)
{
	StackItem *im = *stackp;

	int count = 0;
	Array *ary;

	// Find the boundary item.
	while (im && im->type != STACKITEM_ARY_BOUNDARY) {
		count++;
		im = im->next;
	}

	if (!im) 
		return LAPIS_ERR_ARRAYBOUND;

	StackItem *boundary = im;
	ary = Array_new (count);
	if (count) {
		im = *stackp;
		while (count && im != boundary) {
			StackItem *next = im->next;
			im->next = NULL;
			ary->values[--count] = im;
			im = next;
		}
	}

	boundary->type = STACKITEM_ARRAY;
	boundary->datum.array = ary;
	ary->refcount = 1;
	*stackp = boundary;

	return 0;
}

int cmd_setcookie (short cmdnum, StackItem **stackp, FILE *ouf)
{
	StackItem *tos = *stackp;
	StackItem *nos = tos->next;

	if (tos->type != STACKITEM_STRING ||
	    nos->type != STACKITEM_STRING)
		return LAPIS_ERR_TYPE;

	*stackp = nos->next;
	fprintf (hef, "Set-cookie: %s=%s\n",
		nos->datum.string,
		tos->datum.string);
	StackItem_delete (tos);
	StackItem_delete (nos);
	return 0;
}

int method_Array_load (short cmdnum, StackItem *self, StackItem *in, StackItem **out, FILE *ouf)
{
	Array *a;

	if (self->type != STACKITEM_ARRAY)
		return LAPIS_ERR_TYPE;

	a = self->datum.array;

	// Count the number of provided values.
	int count = 0;
	StackItem *p = in;
	while (p) {
		count++;
		p = p->next;
	}
	if (a->size != count)
		return LAPIS_ERR_LOAD;

	int ix = a->size-1;
	p = in;
	while (p) {
		a->values[ix--] = StackItem_dup (p);
		p = p->next;
	}
	return 0;
}


int method_Matrix_load (short cmdnum, StackItem *self, StackItem *in, StackItem **out, FILE *ouf)
{
	Matrix *m;

	if (self->type != STACKITEM_MATRIX)
		return LAPIS_ERR_TYPE;

	m = self->datum.matrix;

	// Count the number of provided values.
	int count = 0;
	StackItem *p = in;
	while (p) {
		count++;
		p = p->next;
	}
	int size = m->width * m->height;
	if (size != count)
		return LAPIS_ERR_LOAD;

	int ix = size-1;
	p = in;
	while (p) {
		long double value = 0;
		if (p->type == STACKITEM_INTEGER)
			value = (long double) p->datum.integer;
		else
		if (p->type == STACKITEM_REAL)
			value = p->datum.real;
		else
			return LAPIS_ERR_TYPE;

		m->values[ix--] = value;
		p = p->next;
	}
	return 0;
}


int method_Matrix_put (short cmdnum, StackItem *self, StackItem *in, StackItem **out, FILE *ouf)
{
	StackItem *tos = in;
	StackItem *nos = tos->next;
	StackItem *pos = nos->next;

	char type = self->type;

	if (type != STACKITEM_MATRIX)
		return LAPIS_ERR_TYPE;

	if (tos->type != STACKITEM_INTEGER && 
	    tos->type != STACKITEM_REAL) 
		return LAPIS_ERR_TYPE;

	if (nos->type != STACKITEM_INTEGER && 
	    nos->type != STACKITEM_REAL) 
		return LAPIS_ERR_TYPE;

	if (pos->type != STACKITEM_INTEGER && 
	    pos->type != STACKITEM_REAL) 
		return LAPIS_ERR_TYPE;

	int column = pos->type == STACKITEM_INTEGER ? pos->datum.integer : (int)pos->datum.real;
	int row = nos->type == STACKITEM_INTEGER ? nos->datum.integer : (int)nos->datum.real;
	long double value = tos->type == STACKITEM_INTEGER ? (long double)tos->datum.integer : tos->datum.real;

	Matrix *matrix = self->datum.matrix;
	if (matrix) {
		if (row < 0 || row >= matrix->height)
			return LAPIS_ERR_MATRIXROW;
		if (column < 0 || column >= matrix->width)
			return LAPIS_ERR_MATRIXCOLUMN;
		Matrix_put (matrix, row, column, value);
	}

	return 0;
}

int method_Array_put (short cmdnum, StackItem *self, StackItem *in, StackItem **out, FILE *ouf)
{
	StackItem *tos = in;
	StackItem *nos = tos->next;

	char type = self->type;

	if (type != STACKITEM_ARRAY)
		return LAPIS_ERR_TYPE;

	if (nos->type != STACKITEM_INTEGER && 
	    nos->type != STACKITEM_REAL) 
		return LAPIS_ERR_TYPE;

	int ix = nos->type == STACKITEM_INTEGER ?
		nos->datum.integer :
		(int) nos->datum.real ;

	Array *ary = self->datum.array;
	if (ary) {
		if (ix < 0 || ix >= ary->size)
			return LAPIS_ERR_ARRAYINDEX;
		else
			Array_put (ary, ix, StackItem_dup (tos));
	}

	return 0;
}

int method_Array_size (short cmdnum, StackItem *self, StackItem *in, 
		       StackItem **out, FILE *ouf)
{
	if (self->type != STACKITEM_ARRAY)
		return LAPIS_ERR_TYPE;

	Array *a = self->datum.array;
	StackItem *nu = StackItem_new();
	nu->type = STACKITEM_INTEGER;
	nu->datum.integer = a->size;
	*out = nu;
	return 0;
}


int method_Dictionary_count (short cmdnum, StackItem *self, StackItem *in, 
			   StackItem **out, FILE *ouf)
{
	if (self->type != STACKITEM_DICTIONARY)
		return LAPIS_ERR_TYPE;

	Dictionary *d = self->datum.dictionary;
	StackItem *nu = StackItem_new();
	nu->type = STACKITEM_INTEGER;
	int count=0;
	DictionaryItem *im = d->list;
	while (im) {
		count++;
		im = im->next;
	}
	nu->datum.integer = count;
	*out = nu;
	return 0;
}


int method_Dictionary_combine (short cmdnum, StackItem *self, StackItem *in, 
			   StackItem **out, FILE *ouf)
{
	StackItem *nos = self->next;
	char type = self->type;
	Dictionary *primary, *add;
	DictionaryItem *im;

	if (type != STACKITEM_DICTIONARY)
		return LAPIS_ERR_TYPE;
	if (nos->type != STACKITEM_DICTIONARY)
		return LAPIS_ERR_TYPE;

	primary= nos->datum.dictionary;
	add = self->datum.dictionary;

	im = add->list;
	while (im) {
		char *name = im->name;
		StackItem *value = im->value;

		if (!Dictionary_put (primary, name, value))
			return LAPIS_ERR_DATA_RECURSION;

		im = im->next;
	}

	return LAPIS_OK_REMOVE_SELF;
}


int method_Dictionary_put (short cmdnum, StackItem *self, StackItem *in, 
			   StackItem **out, FILE *ouf)
{
	StackItem *tos = in;
	StackItem *nos = tos->next;

	char type = self->type;

	if (type != STACKITEM_DICTIONARY)
		return LAPIS_ERR_TYPE;

	if (nos->type != STACKITEM_SYMBOL && 
	    nos->type != STACKITEM_STRING) 
		return LAPIS_ERR_TYPE;

//printf ("Calling Dict-put, %s := ", nos->datum.string);
//StackItem_print (tos, stdout);
//printf ("\n");
	if (!Dictionary_put (self->datum.dictionary, 
			     nos->datum.string, 
			     StackItem_dup (tos)))
		return LAPIS_ERR_DATA_RECURSION;

	return 0;
}

int method_Dictionary_remove (short cmdnum, StackItem *self, StackItem *in, 
			   StackItem **out, FILE *ouf)
{
	if (self->type != STACKITEM_DICTIONARY)
		return LAPIS_ERR_TYPE;
	if (in->type != STACKITEM_SYMBOL && in->type != STACKITEM_STRING)
		return LAPIS_ERR_TYPE;

	Dictionary_remove (self->datum.dictionary, in->datum.string);

	return 0;
}

int cmd_link (short cmdnum, StackItem **stackp, FILE *ouf)
{
	StackItem *tos = *stackp;
	StackItem *nos = tos->next;
	
	if (tos->type != STACKITEM_STRING ||
	    nos->type != STACKITEM_STRING)
		return LAPIS_ERR_TYPE;

	*stackp = nos->next;
	fprintf (ouf, "<a href=\"%s\">%s</a>\n", 
		nos->datum.string,
		tos->datum.string);

	tos->next = NULL;
	nos->next = NULL;
	StackItem_delete (tos);
	StackItem_delete (nos);
	return 0;
}


int method_Image_new (short cmdnum, StackItem *self, StackItem *in, StackItem **out, FILE *ouf)
{
	StackItem *tos = in;
	StackItem *nos = tos->next;

	if (tos->type != STACKITEM_INTEGER || 
	    nos->type != STACKITEM_INTEGER)
		return LAPIS_ERR_TYPE;

	int width = nos->datum.integer;
	int height = tos->datum.integer;

	StackItem *nu = StackItem_new();

	nu->type = STACKITEM_IMAGE;
	if (!(nu->datum.image = Image_new (width,height)))
		return LAPIS_ERR_NOMEM;
	nu->datum.image->refcount=1;

	*out = nu;

	return 0;
}

int cmd_getcolor (short cmdnum, StackItem **stackp, FILE *ouf)
{
	StackItem *tos = *stackp;

	if (tos->type != STACKITEM_STRING)
		return LAPIS_ERR_TYPE;

	tos->type = STACKITEM_INTEGER;

	unsigned long value=0;
	char *s = tos->datum.string;

	if (!strcasecmp(s, "blue"))
		value=0xff;
	else if (!strcasecmp(s, "red"))
		value=0xff0000;
	else if (!strcasecmp(s, "green"))
		value=0xff00;
	else if (!strcasecmp(s, "white"))
		value=0xffffff;
	else if (!strcasecmp(s, "black"))
		value=0;
	else if (!strcasecmp(s, "cyan"))
		value=0xffff;
	else if (!strcasecmp(s, "maroon"))
		value=0xa03060;
	else if (!strcasecmp(s, "magenta"))
		value=0xff0ff;
	else if (!strcasecmp(s, "yellow"))
		value=0xffff00;
	else if (!strcasecmp(s, "orange"))
		value=0xffa500;
	else if (!strcasecmp(s, "brown"))
		value=0xa52a2a;
	else if (!strcasecmp(s, "purple"))
		value=0xa020f0;
	else if (!strcasecmp(s, "tan"))
		value=0xd0b28a;
	else if (!strcasecmp(s, "navy"))
		value=0x80;

	myfree(s);
	tos->datum.integer = value;
	return 0;
}


int cmd_constant (short cmdnum, StackItem **stackp, FILE *ouf)
{
	StackItem *tos = StackItem_new ();

	switch (cmdnum) {
	case CMD_CONSTANT_PI:
		tos->type = STACKITEM_REAL;
		tos->datum.real = M_PI;
		break;
	case CMD_LPARENS:
		tos->type = STACKITEM_PARAMSTART;
		break;
	case CMD_RPARENS:
		tos->type = STACKITEM_PARAMEND;
		break;
	case CMD_LBRACKET:
		tos->type = STACKITEM_ARY_BOUNDARY;
		break;
	case CMD_CONSTANT_SQRT2:
		tos->type = STACKITEM_REAL;
		tos->datum.real = M_SQRT2;
		break;
	case CMD_CONSTANT_E:
		tos->type = STACKITEM_REAL;
		tos->datum.real = M_E;
		break;
	case CMD_TRUE:
		tos->type = STACKITEM_BOOLEAN;
		tos->datum.integer = 1;
		break;
	case CMD_VERSION:
		tos->datum.string = (uchar*) mystrdup (LAPIS_VERSION);
		tos->type = STACKITEM_STRING;
		break;
	case CMD_NULL:
		tos->type = STACKITEM_NULL;
		tos->datum.string = NULL;
		break;
	case CMD_NOSTYLE:
		tos->type = STACKITEM_STYLE;
		tos->datum.string = NULL;
		break;
	case CMD_FALSE:
		tos->type = STACKITEM_BOOLEAN;
		tos->datum.integer = 0;
		break;
	default:
		return LAPIS_ERR_UNKNOWN;
	}

	tos->next = *stackp;
	*stackp = tos;
	return 0;
}

int method_pow (short cmdnum, StackItem *self, StackItem *in, StackItem **out, FILE *ouf)
{
	char type1, type2;

	type1 = self->type;
	type2 = in->type;

	if (type1 != type2) {
	}
	return 0;
}


int method_binary (short cmdnum, StackItem *self, StackItem *in, StackItem **out, FILE *ouf)
{
	char type1, type2;
	StackItem *nos = self->next;

	if (!nos)
		return LAPIS_ERR_UNDERFLOW;

	type1 = nos->type;
	type2 = self->type;

	if (type1 != type2) {
		// Everything can be converted to a string
		if (type1 == STACKITEM_STRING ||
		    type2 == STACKITEM_STRING) {
			if (type1 != STACKITEM_STRING)
				StackItem_tostring (nos);
			if (type2 != STACKITEM_STRING)
				StackItem_tostring (self);
		}
		else
		// Convert certain things to real if necessary
		if (type1 == STACKITEM_REAL ||
		    type2 == STACKITEM_REAL) {
			if (type1 == STACKITEM_INTEGER) {
				int i = nos->datum.integer;
				nos->type = STACKITEM_REAL;
				nos->datum.real = (long double) i;
			}
			else
			if (type1 == STACKITEM_BOOLEAN) {
				int i = nos->datum.integer;
				nos->type = STACKITEM_REAL;
				nos->datum.real = i ? 1.0 : 0.0;
			} 
			else if (type1 != STACKITEM_REAL)
				return LAPIS_ERR_TYPE;

			if (type2 == STACKITEM_INTEGER) {
				int i = self->datum.integer;
				self->type = STACKITEM_REAL;
				self->datum.real = (long double) i;
			}
			else
			if (type2 == STACKITEM_BOOLEAN) {
				int i = self->datum.integer;
				self->type = STACKITEM_REAL;
				self->datum.real = i ? 1.0 : 0.0;
			} 
			else if (type2 != STACKITEM_REAL)
				return LAPIS_ERR_TYPE;

			nos->type = STACKITEM_REAL;
			self->type = STACKITEM_REAL;
		}
		else
		if (type1 == STACKITEM_BOOLEAN ||
		    type2 == STACKITEM_BOOLEAN) {
			if (type1 == STACKITEM_INTEGER) {
				int i = nos->datum.integer;
				nos->datum.integer = i ? 1 : 0;
			}
			else
			if (type1 == STACKITEM_REAL) {
				nos->datum.integer = nos->datum.real == 0.0 ? 0 : 1;
			} 
			else if (type1 != STACKITEM_BOOLEAN)
				return LAPIS_ERR_TYPE;

			if (type2 == STACKITEM_INTEGER) {
				int i = self->datum.integer;
				self->datum.integer = i ? 1 : 0;
			}
			else
			if (type2 == STACKITEM_REAL) {
				self->datum.integer = nos->datum.real == 0.0 ? 0 : 1;
			} 
			else if (type2 != STACKITEM_BOOLEAN)
				return LAPIS_ERR_TYPE;

			nos->type = STACKITEM_BOOLEAN;
			self->type = STACKITEM_BOOLEAN;
		}
		else
		{
			if (type1 != STACKITEM_INTEGER && type1 != STACKITEM_BOOLEAN) {
				return LAPIS_ERR_TYPE;
			}
				
			if (type2 != STACKITEM_INTEGER && type2 != STACKITEM_BOOLEAN) {
				return LAPIS_ERR_TYPE;
			}
		}
	}
	
	if (self->type == STACKITEM_STRING) {
		char *tempstring;
		int temp;

		switch (cmdnum) {
		case CMD_ADD:
			temp = strlen (self->datum.string);
			temp += strlen (nos->datum.string);
			tempstring = mymalloc (temp+2);
			strcpy (tempstring, nos->datum.string);
			strcat (tempstring, self->datum.string);
//printf ("combined='%s'\n", tempstring);
			if (!tempstring) {
#ifndef TESTING
				log_store ();
				log_writeln ("out of memory");
				log_store ();
#endif
				return myexit(0);
			}
			myfree (nos->datum.string);
			nos->datum.string = tempstring;
			return LAPIS_OK_REMOVE_SELF;
		default:
			StackItem_delete(in);
			return LAPIS_ERR_UNKNOWN;
		}
		return 0;
	}

	if (self->type == STACKITEM_INTEGER)
	{
intversion:
		switch (cmdnum) {
		case CMD_EQUAL: 
			nos->datum.integer =
			 nos->datum.integer == self->datum.integer; 
			nos->type = STACKITEM_BOOLEAN;
			break;
		case CMD_NOTEQUAL: nos->datum.integer =
			nos->datum.integer != self->datum.integer; 
			nos->type = STACKITEM_BOOLEAN;
			break;
		case CMD_ADD:
			nos->datum.integer += self->datum.integer; break;
		case CMD_SUB: 
			nos->datum.integer -= self->datum.integer; break;
		case CMD_MULT: 
			nos->datum.integer *= self->datum.integer; break;
		case CMD_POW: 
			nos->datum.integer = (int) pow (nos->datum.integer, self->datum.integer); break;

		case CMD_BOOLAND: 
			nos->datum.integer = 
				self->datum.integer &&
				nos->datum.integer;
			nos->type = STACKITEM_BOOLEAN;
			break;
		case CMD_AND: 
				nos->datum.integer &= self->datum.integer; 
			break;
		case CMD_BOOLOR: 
			nos->datum.integer =
				self->datum.integer ||
				nos->datum.integer;
			nos->type = STACKITEM_BOOLEAN;
			break;
		case CMD_OR: 
			nos->datum.integer |= self->datum.integer; 
			break;
		case CMD_BOOLXOR:
			nos->datum.integer =
				!!self->datum.integer ^
				!!nos->datum.integer;
			nos->type = STACKITEM_BOOLEAN;
			break;
		case CMD_XOR:
			nos->datum.integer ^= self->datum.integer; 
			break;
		case CMD_LSHIFT:
			nos->datum.integer <<= self->datum.integer; 
			break;
		case CMD_LE:
			nos->datum.integer =
			 nos->datum.integer <= self->datum.integer; 
			nos->type = STACKITEM_BOOLEAN;
			break;
		case CMD_LT:
			nos->datum.integer =
			 nos->datum.integer < self->datum.integer; 
			nos->type = STACKITEM_BOOLEAN;
			break;
		case CMD_RSHIFT:
			nos->datum.integer >>= self->datum.integer; 
			break;
		case CMD_GE:
			nos->datum.integer =
			 nos->datum.integer >= self->datum.integer; 
			nos->type = STACKITEM_BOOLEAN;
			break;
		case CMD_GT:
			nos->datum.integer =
			 nos->datum.integer > self->datum.integer; 
			nos->type = STACKITEM_BOOLEAN;
			break;
		case CMD_DIV:
			if (self->datum.integer) 
				nos->datum.integer /= self->datum.integer; 
			else
				return LAPIS_ERR_DIVIDE;
			break;
		case CMD_REMAINDER:
			if (self->datum.integer) 
				nos->datum.integer %= self->datum.integer; 
			else
				return LAPIS_ERR_DIVIDE;
			break;
		}
	} else 
	if (self->type == STACKITEM_REAL) {
		switch (cmdnum) {
		case CMD_EQUAL: 
			nos->datum.integer =
			nos->datum.real == self->datum.real; 
			nos->type = STACKITEM_BOOLEAN;
			break;
		case CMD_NOTEQUAL: 
			nos->datum.integer =
			 nos->datum.real != self->datum.real; 
			nos->type = STACKITEM_BOOLEAN;
			break;
		case CMD_ADD: nos->datum.real += self->datum.real; break;
		case CMD_SUB: nos->datum.real -= self->datum.real; break;
		case CMD_MULT: nos->datum.real *= self->datum.real; break;
		case CMD_POW: nos->datum.real = pow (nos->datum.real, self->datum.real); break;

		case CMD_LE:
			nos->datum.integer =
			  nos->datum.real <= self->datum.real; 
			nos->type = STACKITEM_BOOLEAN;
			break;
		case CMD_LT:
			nos->datum.integer =
			  nos->datum.real < self->datum.real; 
			nos->type = STACKITEM_BOOLEAN;
			break;
			
		case CMD_GE:
			nos->datum.integer =
			  nos->datum.real >= self->datum.real; 
			nos->type = STACKITEM_BOOLEAN;
			break;

		case CMD_GT:
			nos->datum.integer =
			  nos->datum.real > self->datum.real; 
			nos->type = STACKITEM_BOOLEAN;
			break;

		case CMD_DIV:
			if (self->datum.real) 
				nos->datum.real /= self->datum.real; 
			else
				return LAPIS_ERR_DIVIDE;
			break;

		case CMD_RSHIFT:
		case CMD_LSHIFT:
		case CMD_AND:
		case CMD_OR:
		case CMD_XOR:
		case CMD_REMAINDER:
			/* Convert to integer */
			self->datum.integer = self->datum.real;
			nos->datum.integer = nos->datum.real;
			self->type = nos->type =
				STACKITEM_INTEGER;
			goto intversion;
		}
	}
	else 
	if (self->type == STACKITEM_BOOLEAN) {
		switch (cmdnum) {
		case CMD_AND: 
		case CMD_BOOLAND: 
			nos->datum.integer &= self->datum.integer;
			break;
		case CMD_SUB: 
			if (self->datum.integer)
				nos->datum.integer = 0;
			break;
		case CMD_DIV: 
			if (!self->datum.integer)
				nos->datum.integer = 1;
			break;
		case CMD_ADD: 
		case CMD_OR: 
		case CMD_BOOLOR: 
			nos->datum.integer &= self->datum.integer;
			break;
		case CMD_XOR: 
		case CMD_BOOLXOR: 
		case CMD_MULT: 
			nos->datum.integer ^= self->datum.integer;
			break;
		case CMD_EQUAL: 
			nos->datum.integer = nos->datum.integer == self->datum.integer;
			break;
		case CMD_NOTEQUAL: 
			nos->datum.integer = nos->datum.integer != self->datum.integer;
			break;
		default:
			return LAPIS_ERR_TYPE;
		}
	} else
		return LAPIS_ERR_TYPE;

	return LAPIS_OK_REMOVE_SELF;
}

int method_Array_get (short cmdnum, StackItem *self, StackItem *in, StackItem **out, FILE *ouf)
{
	StackItem *tos;

	tos = in;

	if (self->type != STACKITEM_ARRAY)
		return LAPIS_ERR_TYPE;

	StackItem *value;

	int index = -1;
	int err = 0;

	switch (tos->type) {
	case STACKITEM_INTEGER:
		index = tos->datum.integer; 
		break;
	case STACKITEM_REAL:
		index = (int) tos->datum.real;
		break;
	case STACKITEM_BOOLEAN:
		index = tos->datum.integer & 1; 
		break;
	default:
		err = LAPIS_ERR_TYPE;
	}

	if (err)
		return err;

	if (index >= 0 &&
	    index < self->datum.array->size)
		value = self->datum.array->values[index];
	else
		return LAPIS_ERR_ARRAYINDEX;

	tos->next = NULL;
	StackItem_delete (tos);

	if (value) {
		tos = StackItem_dup (value);
	} else {
		tos = StackItem_new ();
		tos->type = STACKITEM_NULL;
	}

	return 0;
}

int method_Dictionary_get (short cmdnum, StackItem *self, StackItem *in, StackItem **out, FILE *ouf)
{
	StackItem *tos;

	tos = in;
	if (self->type != STACKITEM_DICTIONARY)
		return LAPIS_ERR_TYPE;

	StackItem *value;

	if (tos->type != STACKITEM_SYMBOL && 
	    tos->type != STACKITEM_STRING) 
		return LAPIS_ERR_TYPE;

	value = Dictionary_get (self->datum.dictionary, 
			tos->datum.string);

	if (value) {
		tos = StackItem_dup (value);
	} else {
		tos = StackItem_new ();
		tos->type = STACKITEM_NULL;
	}

	tos->next = *out;
	*out = tos;

	return 0;
}


int cmd_getenv (short cmdnum, StackItem **stackp, FILE *ouf)
{
	StackItem *tos = *stackp;

	if (tos->type != STACKITEM_STRING)
		return LAPIS_ERR_TYPE;

	char *value = getenv (tos->datum.string);
	if (value)
		tos->datum.string = (uchar*) mystrdup (value);
	else
		tos->datum.string = (uchar*) mystrdup ("");
	return 0;
}


int cmd_def (short cmdnum, StackItem **stackp, FILE *ouf)
{
	StackItem *tos = *stackp;
	StackItem *nos = tos->next;

	if (nos->type != STACKITEM_SYMBOL)
		return LAPIS_ERR_TYPE;

	if (!Dictionary_put (global_vars, nos->datum.string, tos))
		return LAPIS_ERR_DATA_RECURSION;

	*stackp = nos->next;
	nos->next = NULL;
	StackItem_delete (nos);
	return 0;
}


int method_unary (short cmdnum, StackItem *self, StackItem *in, StackItem **out, FILE *ouf)
{
	// Only do higher math functions with reals.
	if (self->type == STACKITEM_INTEGER &&
	    (cmdnum == CMD_SIN || cmdnum == CMD_COS || cmdnum == CMD_TAN ||
	     cmdnum == CMD_ATAN || cmdnum == CMD_ASIN || cmdnum == CMD_ACOS))
	{
		long double r = self->datum.integer;
		self->type = STACKITEM_REAL;
		self->datum.real = r;
	}

	if (self->type == STACKITEM_REAL && cmdnum == CMD_NOT) {
		int i = self->datum.real;
		self->type = STACKITEM_BOOLEAN;
		self->datum.integer = !!i;
	}

	int i = self->datum.integer;
	long double r = self->datum.real;

	if (self->type == STACKITEM_INTEGER) {
	}

	int yes=true;

	switch (self->type) {
	case STACKITEM_BOOLEAN:
		if (cmdnum == CMD_NOT)
			 self->datum.integer = !i;
		else
			return LAPIS_ERR_TYPE;
		break;

	case STACKITEM_INTEGER:
		switch (cmdnum) {
		case CMD_ASIN: r = asin((long double)i); break;
		case CMD_ACOS: r = acos((long double)i); break;
		case CMD_ATAN: r = atan((long double)i); break;
		case CMD_SIN: r = sin((long double)i); break;
		case CMD_COS: r = cos((long double)i); break;
		case CMD_TAN: r = tan((long double)i); break;
		case CMD_LN: r = log((long double)i); break;
		default:
			yes=false;
		}
		if (yes) {
			self->datum.real = r;;
			self->type = STACKITEM_REAL;
			return 0;
		}

		switch (cmdnum) {
		case CMD_INCREMENT: i++; break;
		case CMD_NOT: i = !i; break;
		case CMD_INVERT: i = ~i; break;
		case CMD_DECREMENT: i--; break;
		case CMD_NEG: i = -i; break;
		case CMD_SQRT: i = (int) sqrt (i);
		case CMD_CBRT: i = (int) cbrt (i);
		default:
			return LAPIS_ERR_TYPE;
		}
		self->datum.integer = i;
		break;

	case STACKITEM_REAL: 
		switch (cmdnum) {
		case CMD_INCREMENT: r += 1.0; break;
		case CMD_DECREMENT: r -= 1.0; break;
		case CMD_NEG: r = -r; break;
		case CMD_LN: r = log(r); break;
		case CMD_ASIN: r = asin(r); break;
		case CMD_ACOS: r = acos(r); break;
		case CMD_ATAN: r = atan(r); break;
		case CMD_SIN: r = sin(r); break;
		case CMD_COS: r = cos(r); break;
		case CMD_TAN: r = tan(r); break;
		case CMD_SQRT: r = sqrt(r >= 0.0 ? r : 0.0); break;
		case CMD_CBRT: r = cbrt(r >= 0.0 ? r : 0.0); break;
		default:
			return LAPIS_ERR_TYPE;
		}
		self->datum.real = r;
	} 
	return 0;
}



int cmd_print (short cmdnum, StackItem **stackp, FILE *ouf)
{
	StackItem *tos = *stackp;

	if (!content || content == 'h') {
		StackItem_print (tos, ouf);

		if (cmdnum == CMD_WRITELN)
			fprintf (ouf, "\n");
	} else
	if (content == 'j') {
		if (tos->type == STACKITEM_IMAGE) {
			Image *img = tos->datum.image;

			jpeg_image_width = img->width;
			jpeg_image_height = img->height;
			jpeg_image_data = img->pixels;

			char path[30];
			maketempfile (path);
			write_JPEG_file (path, 99);

			fclose (ouf);
			ouf = NULL;

			if (!outpath2)
				return LAPIS_ERR_INTERNAL;

			unlink (outpath2);
			if (rename (path, outpath2)) {
				perror ("rename");
			}
		}
		else
			return LAPIS_ERR_TYPE;
	}

	tos->next = NULL;
	*stackp = tos->next;
	StackItem_delete(tos);

	return 0;
}


int cmd_stackop (short cmdnum, StackItem **stackp, FILE *ouf)
{
	StackItem *tos = *stackp;
	StackItem *nos = tos->next;

	switch (cmdnum) {
	case CMD_2DUP: {
		StackItem *tos2 = StackItem_dup (tos);
		StackItem *nos2 = StackItem_dup (nos);
		tos2->next = nos2;
		nos2->next = tos;
		*stackp = tos2;
		}
		break;
	case CMD_2DROP: {
		StackItem *nos = tos->next;
		*stackp = nos->next;
		tos->next = NULL;
		nos->next = NULL;
		StackItem_delete (tos);
		StackItem_delete (nos);
		}
		break;
	case CMD_SWAP:
	case CMD_EXCH:
		*stackp = nos;
		tos->next = nos->next;
		nos->next = tos;
		break;
	case CMD_DUP: {
		StackItem *tos2 = StackItem_dup (tos);
		tos2->next = tos;
		*stackp = tos2;
		}
		break;
	case CMD_DROP:
		*stackp = tos->next;
		tos->next = NULL;
		StackItem_delete (tos);
		break;
	case CMD_NIP:
		tos->next = nos->next;
		nos->next = NULL;
		StackItem_delete (nos);
		break;
	case CMD_OVER: {
		StackItem *im = StackItem_dup (nos);
		im->next = tos;
		*stackp = im;
		}
		break;
	case CMD_ROT: {
		StackItem *n3 = nos->next;
		StackItem *rest = n3->next;
		*stackp = n3;
		n3->next = tos;
		nos->next = rest;
		}
		break;
	default:
		return LAPIS_ERR_UNKNOWN;
	}
	return 0;
}


int method_Matrix_new (short cmdnum, StackItem *self, StackItem *in, StackItem **out, FILE *ouf)
{
	StackItem *tos = in;
	StackItem *nos = tos->next;

	if (tos->type != STACKITEM_INTEGER || 
	    nos->type != STACKITEM_INTEGER)
		return LAPIS_ERR_TYPE;

	StackItem *nu = StackItem_new ();

	int width = nos->datum.integer;
	int height = tos->datum.integer;

	nu->type = STACKITEM_MATRIX;
	if (!(nu->datum.matrix = Matrix_new (width,height)))
		return LAPIS_ERR_NOMEM;
	nu->datum.matrix->refcount=1;

	*out = nu;

	return 0;
}


int method_Array_new (short cmdnum, StackItem *self, StackItem *in, StackItem **out, FILE *ouf)
{
	StackItem *tos = in;

	if (tos->type == STACKITEM_REAL) {
		long double d = tos->datum.real;
		tos->datum.integer = d;
		tos->type = STACKITEM_INTEGER;
	}

	if (tos->type != STACKITEM_INTEGER)
		return LAPIS_ERR_TYPE;

	int initial_size = tos->datum.integer;

	StackItem *nu = StackItem_new();

	nu->type = STACKITEM_ARRAY;
	if (!(nu->datum.array = Array_new (initial_size))) 
		return LAPIS_ERR_NOMEM;

	nu->datum.array->refcount=1;
	
	*out = nu;

	return 0;
}

int method_Dictionary_new (short cmdnum, StackItem *self, StackItem *in, StackItem **out, FILE *ouf)
{
	StackItem *nu = StackItem_new();

	nu->type = STACKITEM_DICTIONARY;
	if (!(nu->datum.dictionary = Dictionary_new ())) 
		return LAPIS_ERR_NOMEM;

	nu->datum.dictionary->refcount=1;
	
	*out = nu;

	return 0;
}

int method_Table_new (short cmdnum, StackItem *self, StackItem *in, StackItem **out, FILE *ouf)
{
	StackItem *nu = StackItem_new();

	nu->type = STACKITEM_TABLE;
	if (!(nu->datum.table = Table_new ())) 
		return LAPIS_ERR_NOMEM;

	nu->datum.table->refcount=1;
	
	*out = nu;

	return 0;
}

int cmd_content (short cmdnum, StackItem **stackp, FILE *ouf)
{
	switch (cmdnum) {
	case CMD_CONTENT_HTML:
		fprintf (hef, "Content-type: text/html\n\n");
		content = 'h';
		break;
	case CMD_CONTENT_JPEG: 	
		fprintf (hef, "Content-type: image/jpeg\n\n");
		content = 'j';
		break;
	}
	return 0;
}

int method_Image_graphic (short cmdnum, StackItem *self, StackItem *in, StackItem **out, FILE *ouf)
{
	StackItem *tos = in;
	StackItem *nos = tos->next;

	short x,y;

	if (self->type != STACKITEM_IMAGE)
		return LAPIS_ERR_TYPE;

	if (tos->type == STACKITEM_INTEGER)
		y = tos->datum.integer;
	else
	if (tos->type == STACKITEM_REAL)
		y = tos->datum.real;
	else
		return LAPIS_ERR_TYPE;

	if (nos->type == STACKITEM_INTEGER)
		x = nos->datum.integer;
	else
	if (nos->type == STACKITEM_REAL)
		x = nos->datum.real;
	else
		return LAPIS_ERR_TYPE;

	switch (cmdnum) {
	case CMD_MOVETO: 
		self->datum.image->x = x;
		self->datum.image->y = y;
		break;
	case CMD_LINETO:
		Image_lineto (self->datum.image, x, y);
		break;
	case CMD_FILLRECT:
		Image_fillrect (self->datum.image, x, y);
		break;
	}

	return 0;
}


int cmd_body (short cmdnum, StackItem **stackp, FILE *ouf)
{
	StackItem *tos = *stackp;
	StackItem *nos = tos->next;
	StackItem *pos = nos->next;

	if (tos->type != STACKITEM_INTEGER ||
	    nos->type != STACKITEM_INTEGER ||
	    pos->type != STACKITEM_INTEGER)
		return LAPIS_ERR_TYPE;

	*stackp = pos->next;

	fprintf (ouf, "\n<body bgcolor=#%06x link=#%06x vlink=#%06x>\n",
		pos->datum.integer,
		nos->datum.integer,
		tos->datum.integer);

	tos->next = NULL;
	nos->next = NULL;
	pos->next = NULL;
	StackItem_delete (tos);
	StackItem_delete (nos);
	StackItem_delete (pos);
	return 0;
}


// Example: 11.0 "Times" font-begin

int cmd_fontbegin (short cmdnum, StackItem **stackp, FILE *ouf)
{
	StackItem *tos = *stackp;
	StackItem *nos;

	if (tos->type != STACKITEM_STRING) 
		return LAPIS_ERR_TYPE;
	nos = tos->next;
	if (nos->type != STACKITEM_INTEGER && 
	    nos->type != STACKITEM_REAL) 
		return LAPIS_ERR_TYPE;
	*stackp = nos->next;
	if (nos->type == STACKITEM_REAL)
		fprintf (ouf, "<span style=\"font: %Lfpt %s;\">", nos->datum.real,
			tos->datum.string);
	else
		fprintf (ouf, "<span style=\"font: %dpt %s;\">", nos->datum.integer,
			tos->datum.string);

	tos->next = NULL;
	nos->next = NULL;
	StackItem_delete (tos);
	StackItem_delete (nos);
	return 0;
}

int cmd_fontface (short cmdnum, StackItem **stackp, FILE *ouf)
{
	StackItem *tos = *stackp;
	if (tos->type != STACKITEM_STRING)
		return LAPIS_ERR_TYPE;
	*stackp = tos->next;

	fprintf (ouf, "<font face=\"%s\">", tos->datum.string);
	tos->next = NULL;
	StackItem_delete (tos);
	return 0;
}

int cmd_fontsize (short cmdnum, StackItem **stackp, FILE *ouf)
{
	StackItem *tos = *stackp;

	if (tos->type != STACKITEM_INTEGER)
		return LAPIS_ERR_TYPE;

	*stackp = tos->next;

	fprintf (ouf, "<span style=\"font-size: %dpt;\">", tos->datum.integer);
	tos->next = NULL;
	StackItem_delete (tos);

	return 0;
}


int method_Font_getconsolefont (short cmdnum, StackItem *self, StackItem *in, StackItem **out, FILE *ouf)
{
	StackItem *tos = in;

	if (tos->type != STACKITEM_STRING)
		return LAPIS_ERR_TYPE;

	Font *font = Font_load_consolefont (tos->datum.string);
	if (font) {
		myfree (tos->datum.string);	
		tos->type = STACKITEM_FONT;
		tos->datum.font = font;
		font->refcount++;
	} else {
		tos->type = STACKITEM_NULL;
	}
	return 0;
}


/* These are commands that cannot be overridden
 * in any class.
 */
static struct cmdinfo commands_nonalpha [] = {
	{ "=",   0, 0, cmd_def, NULL , NULL, 0, 0 },
	{ "(",   CMD_LPARENS, 0, cmd_constant, NULL , NULL, 0, 0 },
	{ ")",   CMD_RPARENS, 1, cmd_constant, NULL , NULL, 0, 0 },
	{ "{",   CMD_LBRACE, 0, cmd_lbrace, NULL , NULL, 0, 0 },
	{ "}",   CMD_RBRACE, 1, cmd_rbrace, NULL , NULL, 0, 0 },
	{ "[",   CMD_LBRACKET, 0, cmd_constant, NULL , NULL, 0, 0 },
	{ "]",   CMD_RBRACKET, 1, cmd_closearray, NULL , NULL, 0, 0 },
	{ "2dup",   CMD_2DUP, 2, cmd_stackop, NULL , NULL, 0, 0 },
	{ "2drop",  CMD_2DROP, 2, cmd_stackop, NULL , NULL, 0, 0 },
	{ NULL, 0, 0, NULL, NULL, NULL, 0, 0},
};
static struct cmdinfo commands_a [] = {
	{ NULL, 0, 0, NULL, NULL, NULL, 0, 0},
};
static struct cmdinfo commands_b [] = {
	// { "big-begin",   0, NULL, "<big>" , NULL, 0, 0 },
	// { "big-end",   0, NULL, "</big>" , NULL, 0, 0 },
	// { "big",   0, NULL, NULL, NULL, 0, 1 },
	{ "body", 0, 3, cmd_body, NULL , NULL, 0, 0 },
	{ "bold", 0, 1, NULL, NULL , NULL, ATTR_BOLD , 0},
	{ "bold-begin", 0, 0, NULL, "<b>" , NULL, 0, 0 },
	{ "bold-end", 0, 0, NULL, "</b>" , NULL, 0, 0 },
	{ "br", 0, 0, NULL, "<br>\n\n", NULL, 0, 0 },
	{ NULL, 0, 0, NULL, NULL, NULL, 0, 0},
};
static struct cmdinfo commands_c [] = {
	{ "center-begin", 0, 0, NULL, "<center>" , NULL, 0, 0 },
	{ "center-end", 0, 0, NULL, "</center>\n" , NULL, 0, 0 },
	{ "content-html", CMD_CONTENT_HTML, 0, cmd_content, NULL , NULL, 0, 0 },
	{ "content-jpeg", CMD_CONTENT_JPEG, 0, cmd_content, NULL , NULL, 0, 0 },
	{ "constant_e",  CMD_CONSTANT_E, 0, cmd_constant, NULL , NULL, 0, 0 },
	{ NULL, 0, 0, NULL, NULL, NULL, 0, 0},
};
static struct cmdinfo commands_d [] = {
	{ "div-begin", 0,  1, NULL, NULL , "<div" , 0, 0 },
	{ "div-end", 0,  0, NULL, "</div>" , NULL, 0, 0 },
	{ "drop", CMD_DROP,  1, cmd_stackop, NULL , NULL, 0, 0 },
	{ "dup", CMD_DUP,  1, cmd_stackop, NULL , NULL, 0, 0 },
	{ NULL, 0, 0, NULL, NULL, NULL, 0, 0},
};
static struct cmdinfo commands_e [] = {
	{ "exch", CMD_EXCH,  2, cmd_stackop, NULL , NULL, 0, 0 },
	{ "exit", 0,  1, cmd_exit, NULL , NULL, 0, 0 },
	{ NULL, 0, 0, NULL, NULL, NULL, 0, 0},
};
static struct cmdinfo commands_f [] = {
	{ "false", CMD_FALSE,  0, cmd_constant, NULL , NULL, 0, 0 },
	{ "fatal", 0,  1, cmd_fatal, NULL , NULL, 0, 0 },
	{ "font-begin", 0, 0, cmd_fontface, NULL, NULL, 0, 0 },
	{ "font-end", 0, 0, NULL, "</font>", NULL, 0, 0 },
	{ "fontsize-begin", 0, 0, cmd_fontsize, NULL, NULL, 0, 0 },
	{ "fontsize-end", 0, 0, NULL, "</span>", NULL, 0, 0 },
	{ NULL, 0, 0, NULL, NULL, NULL, 0, 0},
};
static struct cmdinfo commands_g [] = {
	{ "getcolor", 0,  1, cmd_getcolor, NULL , NULL, 0, 0 },
	{ "getenv", 0,  1, cmd_getenv, NULL , NULL, 0, 0 },
	{ NULL, 0, 0, NULL, NULL, NULL, 0, 0},
};
static struct cmdinfo commands_h [] = {
	{ "hrule",   0, 0, NULL, "\n<hr>\n" , NULL, 0, 0 },
	{ "h1-begin",0, 0, NULL, "<h1>" , NULL, 0, 0 },
	{ "h1-end",  0, 0, NULL, "</h1>" , NULL, 0, 0 },
	{ "h2-begin",0, 0, NULL, "<h2>" , NULL, 0, 0 },
	{ "h2-end",  0, 0, NULL, "</h2>" , NULL, 0, 0 },
	{ "h3-begin",0, 0, NULL, "<h3>" , NULL, 0, 0 },
	{ "h3-end",  0, 0, NULL, "</h3>" , NULL, 0, 0 },
	{ "h4-begin",0, 0, NULL, "<h4>" , NULL, 0, 0 },
	{ "h4-end",  0, 0, NULL, "</h4>" , NULL, 0, 0 },
	{ "h5-begin",0, 0, NULL, "<h5>" , NULL, 0, 0 },
	{ "h5-end",  0, 0, NULL, "</h5>" , NULL, 0, 0 },
	{ "h6-begin",0, 0, NULL, "<h6>" , NULL, 0, 0 },
	{ "h6-end",  0, 0, NULL, "</h6>" , NULL, 0, 0 },
	{ NULL, 0, 0, NULL, NULL, NULL, 0, 0},
};
static struct cmdinfo commands_i [] = {
	{ "indent-begin", 0,   0, NULL, "<dir>\n" , NULL, 0, 0 },
	{ "indent-end",   0, 0, NULL, "</dir>\n" , NULL, 0, 0 },
	{ "italic",  0,  1, NULL, NULL , NULL, ATTR_ITALIC, 0 },
	{ "italic-begin", 0,   0, NULL, "<i>" , NULL, 0, 0 },
	{ "italic-end",  0,  0, NULL, "</i>" , NULL, 0, 0 },
	{ "item",  0,  0, NULL, "<li>" , NULL, 0, 0 },
	{ NULL, 0,0, NULL, NULL, NULL, 0, 0},
};
static struct cmdinfo commands_l [] = {
	{ "link", 0,  2, cmd_link, NULL , NULL, 0, 0 },
	{ "list-begin", 0,  0, NULL, "<ul>\n" , NULL, 0, 0 },
	{ "list-end", 0,  0, NULL, "</ul>\n" , NULL, 0, 0 },
	{ NULL, 0,0, NULL, NULL, NULL, 0, 0},
};
static struct cmdinfo commands_m [] = {
	{ NULL, 0,0, NULL, NULL, NULL, 0, 0},
};
static struct cmdinfo commands_n [] = {
	{ "nip", 0,  2, cmd_stackop, NULL , NULL, 0, 0 },
	{ "nostyle", CMD_NOSTYLE,  0, cmd_constant, NULL , NULL, 0, 0 },
	{ "null", CMD_NULL,  0, cmd_constant, NULL , NULL, 0, 0 },
	{ "numbered-begin",0, 0, NULL, "<ol>\n" , NULL, 0, 0 },
	{ "numbered-end",  0, 0, NULL, "</ol>\n" , NULL, 0, 0 },
	{ NULL, 0, 0, NULL, NULL, NULL, 0, 0},
};
static struct cmdinfo commands_o [] = {
	{ "over", CMD_OVER,  2, cmd_stackop, NULL , NULL, 0, 0 },
	{ NULL, 0, 0, NULL, NULL, NULL, 0, 0},
};
static struct cmdinfo commands_p [] = {
	{ "par", 0,  0, NULL, "<p>\n" , NULL, 0, 0 },
	{ "para-begin", 0,1, NULL, NULL , "<p" , 0, 0 },
	{ "para-end",   0,0, NULL, "</p>\n" , NULL, 0, 0 },
	{ "pi", CMD_CONSTANT_PI, 0, cmd_constant, NULL , NULL, 0, 0 },
	{ "plain", 0, 0, NULL, NULL , NULL, ATTR_PLAIN, 0 },
	// { "print", 0,  1, cmd_print, NULL , NULL, 0, 0 },
	{ NULL, 0,0, NULL, NULL, NULL, 0, 0},
};
static struct cmdinfo commands_r [] = {
	{ "right-begin", 0,0, NULL, "<p align=right>" , NULL, 0, 0 },
	{ "right-end",   0,0, NULL, "</p>" , NULL, 0, 0 },
	{ "rot",  CMD_ROT, 3, cmd_stackop, NULL , NULL, 0, 0 },
	{ NULL, 0, 0, NULL, NULL, NULL, 0, 0},
};
static struct cmdinfo commands_s [] = {
	{ "set-cookie", 0,  2, cmd_setcookie, NULL , NULL, 0, 0 },
	{ "small-begin", 0,  0, NULL, "<small>" , NULL, 0, 0 },
	{ "small-end", 0,  0, NULL, "</small>" , NULL, 0, 0 },
	{ "small", 0,  0, NULL, NULL, NULL, 0, -1 },
	{ "span-begin", 0,  1, NULL, NULL , "<span" , 0, 0 },
	{ "span-end", 0,  0, NULL, "</span>" , NULL , 0, 0 },
	{ "strike-begin", 0, 0, NULL, "<s>" , NULL, 0, 0 },
	{ "strike-end", 0, 0, NULL, "</s>" , NULL, 0, 0 },
	{ "strike", 0, 1, NULL, NULL , NULL, ATTR_STRIKEOUT, 0},
	{ "style", 0,  1, cmd_style, NULL , NULL, 0, 0 },
	{ "subscript", 0,  1, NULL, NULL , NULL, ATTR_SUBSCRIPT, 0 },
	{ "superscript",0,   1, NULL, NULL , NULL , ATTR_SUPERSCRIPT, 0},
	{ "swap", CMD_SWAP,  2, cmd_stackop, NULL , NULL, 0, 0 },
	{ "sqrt2", CMD_CONSTANT_SQRT2,  0, cmd_constant, NULL , NULL, 0, 0 },
	{ NULL, 0, 0, NULL, NULL, NULL, 0, 0},
};
static struct cmdinfo commands_t [] = {
	{ "table-begin", 0,  1, NULL, NULL , "<table" , 0, 0 },
	{ "table-column", 0,  1, NULL, NULL , "<td" , 0, 0 },
	{ "table-row", 0,  1, NULL, NULL, "<tr" , 0, 0 },
	{ "table-end", 0,  0, NULL, "</table>\n" , NULL, 0, 0 },
	{ "table-hdr", 0,  0, NULL, "<th>\n" , NULL, 0, 0 },
	{ "title", 0,  1, NULL, NULL , NULL, ATTR_TITLE, 0 },
	{ "true", CMD_TRUE,  0, cmd_constant, NULL , NULL, 0, 0 },
	{ NULL, 0, 0, NULL, NULL, NULL, 0, 0},
};
static struct cmdinfo commands_u [] = {
	{ "underline", 0,  1, NULL, NULL , NULL, 2, 0 },
	{ "underline-begin", 0,  0, NULL, "<ul>" , NULL, 0, 0 },
	{ "underline-end", 0,  0, NULL, "</ul>" , NULL, 0, 0 },
	{ NULL, 0,0, NULL, NULL, NULL, 0, 0},
};
static struct cmdinfo commands_v [] = {
	{ "version", CMD_VERSION,  0, cmd_constant, NULL , NULL, 0, 0 },
	{ NULL, 0, 0, NULL, NULL, NULL, 0, 0},
};
static struct cmdinfo commands_w [] = {
	{ NULL, 0, 0, NULL, NULL, NULL, 0, 0},
};


static struct cmdinfo *commands[27] = {
	commands_a,
	commands_b,
	commands_c,
	commands_d,
	commands_e,
	commands_f,
	commands_g,
	commands_h,
	commands_i,
	NULL,
	NULL,
	commands_l,
	commands_m,
	commands_n,
	commands_o,
	commands_p,
	NULL,
	commands_r,
	commands_s,
	commands_t,
	commands_u,
	commands_v,
	commands_w,
	NULL,
	NULL,
	NULL,
	commands_nonalpha,
};


struct cmdinfo *lookup_command (char *s)
{
	int ch = tolower(*s);
	int ix = isalpha(ch) ? ch-'a' : 26;
	struct cmdinfo *ptr;
	ptr = commands[ix];
	while (ptr && ptr->name) {
		if (!strcmp (s, ptr->name))
			return ptr;
		ptr++;
	}
	return NULL;
}

Class *lookup_class (char *name);
Class *StackItem_getclass (StackItem *im);


bool process_lapis_core (int fd, char *codepath, char *inpath, char *outpath,
	char *headerpath)
{
	FILE *inf, *ouf, *f;
	StackItem *stack = NULL;

	if (!inpath || !outpath)
		return false;
	/*---------*/

	f = fopen(codepath, "rb");
	if (!f) {
		perror ("fopen read");
		return false;
	}
	inf = fopen(inpath, "rb");
	if (!inf) {
		fclose(f);
		perror ("fopen read");
		return false;
	}
	ouf = fopen(outpath, "wb");
	if (!ouf) {
		perror ("fopen write");
		fclose(f);
		fclose(inf);
		return false;
	}
	hef = fopen(headerpath, "wb"); // hef is global
	if (!hef) {
		perror ("fopen write");
		fclose(f);
		fclose(inf);
		fclose(ouf);
		return false;
	}

	outpath2 = outpath; // needed by "print" command

	char word[100];
	int len;
	char err=0;

	while (!err && (len = getword (f, word, 99)) > 0) {
		char ch, ch2;

//printf ("WORD '%s'\n", word);
		ch = word[0];
		ch2 = word[1];

		// Handle the pushing of HEX constants
		//
		if (ch=='0' && ch2=='x') {
			unsigned long value;
			if (1 == sscanf (word+2, "%lx", &value)) {
				StackItem *im = StackItem_new();
				im->datum.integer = value;
				im->type = STACKITEM_INTEGER;
				im->next = stack;
				stack = im;
			}
			continue;
		} 
		else
		// Handle the pushing of integer and real constants
		//
		if (isdigit (ch) || ('-' == ch && isdigit(ch2))) {
			StackItem *im = StackItem_new();
			if (!im) {
				err = LAPIS_ERR_NOMEM;
				break;
			}
//printf ("new item = $%08lx\n", (unsigned long)im);
			if (strchr (word, '.')) {
				long double real;
				if (1 == sscanf (word, "%Lf", &real)) {
					im->type = STACKITEM_REAL;
					im->datum.real = real;
				} else {
					StackItem_delete (im);
					err = LAPIS_ERR_BADREAL;
					break;
				}
				im->next = stack;
				stack = im;
			} else {
				int integer;
				if (1 == sscanf (word, "%d", &integer)) {
					im->type = STACKITEM_INTEGER;
					im->datum.integer = integer;
				} else
					break;
				im->next = stack;
				stack = im;
			}
			continue;
		}
		else
		// Handle the pushing of symbols
		//
		if ('/' == ch && word[1]) {
#if 0
			if (strchr (word+1, '\''))
				break;
#endif

//printf ("SYMBOL '%s'\n", word);
			// Make sure it isn't a reserved word.
			struct cmdinfo *cmd = lookup_command (&word[1]);
			if (cmd) {
				err = LAPIS_ERR_RESERVED;
				break;
			}

			StackItem *im = StackItem_new();
			im->type = STACKITEM_SYMBOL;
			im->datum.string = (uchar*) mystrdup(word+1);
			im->next = stack;
			stack = im;
			continue;
		}
		else
		// Handle the pushing of strings
		//
		if (word[0] == '"') {
			StackItem *im = StackItem_new();
			im->datum.string = (uchar*) mystrdup (word+1);
//printf ("string allocated as $%08lx\n", (unsigned long)im->datum.string);
			im->type = STACKITEM_STRING;
			im->next = stack;
			stack = im;
			continue;
		}
		else
		// Handle the deletion of global variables
		//
		if (word[0] == '~' && isalpha(word[1])) {
			if (!Dictionary_remove (global_vars, word+1))
				err = LAPIS_ERR_VARIABLE;
			continue;
		}
		else
		{
			Class *class=NULL;
			Method *method=NULL;
			StackItem *im=NULL;

			// Count the # of params, up to 8.
			int nparam = 0;
			im = stack;
			while (nparam < 8 && im) {
				char type = im->type;
				if (type != STACKITEM_PARAMSTART && type != STACKITEM_PARAMEND)
					nparam++;
				im = im->next;
			}

			// Is it a non-OOP command e.g. getenv?
			struct cmdinfo *info = lookup_command (word);
			if (info) {
				if (info->flags) {
					unsigned char f = info->flags;
					if (!stack) {
						err = LAPIS_ERR_UNDERFLOW;
						break;
					}
					if (f & ATTR_BOLD)
						stack->bold = true;
					else
					if (f & ATTR_ITALIC)
						stack->italic = true;
					else
					if (f & ATTR_UNDERLINE)
						stack->underline = true;
					else
					if (f & ATTR_STRIKEOUT)
						stack->strikeout = true;
					else
					if (f & ATTR_SUPERSCRIPT)
						stack->super = true;
					else
					if (f & ATTR_SUBSCRIPT)
						stack->sub = true;
					else
					if (f & ATTR_PLAIN) {
						stack->bold = false;
						stack->italic = false;
						stack->underline = false;
						stack->super = false;
						stack->sub = false;
						stack->strikeout = false;
					}
					continue;
				}
				else
				if (info->func) {
					if (nparam < info->num_params) {
//printf ("cmdinfo says we need %d params\n", info->num_params);
						err = LAPIS_ERR_UNDERFLOW;
						break;
					}

					err = info->func (info->cmdnum, &stack, ouf);
					if (err)
						break;
					continue;
				} 
				else
				if (info->to_print) {
					fprintf (ouf, info->to_print);
				}
				else
				if (info->print_with_style) {
					if (!stack) {
						err = LAPIS_ERR_UNDERFLOW;
						break;
					}
					if (stack->type != STACKITEM_STRING &&
					    stack->type != STACKITEM_STYLE)
					{
						err = LAPIS_ERR_TYPE;
						break;
					}
					im = stack;
					stack = stack->next;
					im->next = NULL;
					fprintf (ouf, info->print_with_style);
					fprintf (ouf, " style=\"%s\">", im->datum.string);
					StackItem_delete (im);
					continue;
				}
				continue;
			}

			// Is it a class name?
			// If so, we're creating a new instance.
			if (ch >= 'A' && ch <= 'Z') {
				class = lookup_class (word);
				if (class) {
//printf ("Class constructor %s found\n", word);
					method = Class_get_method (class, "new");
					if (!method) {
						err = LAPIS_ERR_NO_CONSTRUCTOR;
						break;
					}
					if (nparam < method->nparams) {
						err = LAPIS_ERR_UNDERFLOW;
						break;
					}
					StackItem *outstack = NULL;
					StackItem *instack = NULL;
					int i = method->nparams;
					if (i) {
						// Isolate the input stack items
						// which is also safer.
						//
						StackItem *prev=NULL;
						instack = stack;
						while (i--) {
							prev = stack;
							stack = stack->next;
						}
						prev->next = NULL;
					}
					err = method->func (0, NULL, instack, &outstack, ouf);
					if (err)
						break;
					if (outstack) {
						// Push the result stack onto the regular stack.
						StackItem *last=outstack;
						while (last->next) {
							last = last->next;
						}
						last->next = stack;
						stack = outstack;					
					}
					continue;
				}
			}

#if 0
printf ("stack type %d\n", stack?stack->type:-1);
if (stack && stack->type<0)
{
	char *a=0; *a=0;
}
#endif
			// Is it a method that follows a bunch of params?
			//
			if (stack && stack->type == STACKITEM_PARAMEND) {

				// Get rid of the ')'
				im = stack;
				stack = im->next;
				im->next = NULL;
				StackItem_delete (im);

				// Find the object and isolate the stack.
				StackItem *obj = NULL;
				StackItem *param_start = stack;
				StackItem *parens = stack;
				StackItem *above_parens = NULL;

				while (parens && parens->type != STACKITEM_PARAMSTART) {
					above_parens = parens;
					parens = parens->next;
				}
				if (!parens) {
					err = LAPIS_ERR_PARAMBOUND;
					break;
				}
				obj = parens->next;
				parens->next = NULL;
				above_parens->next = NULL;
				StackItem_delete (parens);
				while (obj && obj->type == STACKITEM_PARAMSTART) {
					obj = obj->next;
				}
				if (!obj) {
					err = LAPIS_ERR_NO_OBJ;
					break;
				}

				StackItem *outstack = NULL;

				stack = obj;

				StackItem *self = NULL;
				char *methodstr = NULL;

				// First look for a variable name combined with a method name.
				if ((methodstr = strchr (word, '.'))) {
					*methodstr++ = 0;
					if (strchr (methodstr, '.')) {
						err = LAPIS_ERR_BADVARMETHOD;
						break;
					}

					self = Dictionary_get (global_vars, word);
					if (!self) {
						err = LAPIS_ERR_VARIABLE;
						break;
					}
				}
				else 
				{
					self = obj;
					methodstr = word;
				}

//printf ("method='%s'\n",methodstr);
				Class *class = StackItem_getclass (self);
				if (!class) {
					err = LAPIS_ERR_CANNOT_GET_CLASS;
					break;
				}
				method = Class_get_method (class, methodstr);
				if (!method) {
					err = LAPIS_ERR_UNKNOWN_METHOD;
					break;
				}

				if (method) {
//printf ("Calling %s::%s\n", class->name, word);
					if (method->nparams != -1 &&
					    nparam < method->nparams) {
						err = LAPIS_ERR_UNDERFLOW;
						break;
					}
					err = method->func (method->cmdnum, self, param_start, &outstack, ouf);
					if (obj==self && err == LAPIS_OK_REMOVE_SELF) {
						stack = obj->next;
						obj->next = NULL;
						StackItem_delete (obj);
						err = 0;
					}
					else
					if (err) 
						break;

					if (param_start) {
						// Destroy the input data
						while (param_start) {
							StackItem *next = param_start->next;
							param_start->next = NULL;
							StackItem_delete (param_start);	
							param_start = next;
						}
					}
					if (outstack) {
						// Push the result stack onto the regular stack.
						StackItem *last=outstack;
						while (last->next) {
							last = last->next;
						}
						last->next = stack;
						stack = outstack;					
					}
					continue;
				} 
				err = LAPIS_ERR_UNKNOWN_METHOD;
				break;
			}

			// Is it a no-params-required method for the TOS object?
			//
			Class *tos_class = StackItem_getclass (stack);
			if (tos_class) {
				method = Class_get_method (tos_class, word);
				if (method) {
					if (method->nparams) {
						err = LAPIS_ERR_NEED_PARAMS;
						break;
					}
					StackItem *outstack = NULL;
					err = method->func (method->cmdnum, stack, NULL, &outstack, ouf);
					if (err == LAPIS_OK_REMOVE_SELF) {
						im = stack;
						stack = im->next;
						im->next = NULL;
						StackItem_delete (im);
						err = 0;
					}
					else
					if (err) 
						break;
					if (outstack) {
						// Push the result stack onto the regular stack.
						StackItem *last=outstack;
						while (last->next) {
							last = last->next;
						}
						last->next = stack;
						stack = outstack;					
					}
					continue;
				} 
			}

			// Is it a global variable name?
			im = Dictionary_get (global_vars, word);
			if (im) {
				StackItem *dup = StackItem_dup (im);
				dup->next = stack;
				stack = dup;
				continue;
			}

			err = LAPIS_ERR_UNKNOWN;
			break;
		}
	}

	char *s = NULL;
	switch (err) {
	case LAPIS_ERR_UNDERFLOW:	s = "stack underflow"; break;
	case LAPIS_ERR_DIVIDE: 	s = "divide by zero"; break;
	case LAPIS_ERR_TYPE:	s = "wrong type parameter for this method"; break;
	case LAPIS_ERR_UNKNOWN:	s = "unknown keyword"; break;
	case LAPIS_ERR_NOMEM:	s = "out of memory"; break;
	case LAPIS_ERR_NOFONT:	s = "cannot load font"; break;
	case LAPIS_ERR_PARAMBOUND:	s = "parameter list lacks '('"; break;
	case LAPIS_ERR_ARRAYBOUND:	s = "array definition lacks '['"; break;
	case LAPIS_ERR_NOFUNC:	s = "function not implemented"; break;
	case LAPIS_ERR_INTERNAL:	s = "internal"; break;
	case LAPIS_ERR_RESERVED:	s = "can't use reserved word as symbol"; break;
	case LAPIS_ERR_UNIMPLEMENTED:	s = "feature not implemented"; break;
	case LAPIS_ERR_UNKNOWN_METHOD:	s = "unknown method"; break;
	case LAPIS_ERR_METHOD_WITH_DOT:	s = "method has dot parameter"; break;
	case LAPIS_ERR_CANNOT_GET_CLASS:	s = "cannot determine variable's class"; break;
	case LAPIS_ERR_FATAL:	break;
	case LAPIS_ERR_NO_CONSTRUCTOR:	s = "class lacks constructor"; break;
	case LAPIS_ERR_NEED_PARAMS:	s = "method requires parameters"; break;
	case LAPIS_ERR_NO_OBJ:	s = "method lacks object"; break;
	case LAPIS_OK_REMOVE_SELF:	s = "ok, object removed"; break;
	case LAPIS_ERR_DATA_RECURSION:	s = "datum stored inside itself"; break;
	case LAPIS_ERR_MATRIXROW:	s = "invalid matrix row"; break;
	case LAPIS_ERR_MATRIXCOLUMN:	s = "invalid matrix column"; break;
	case LAPIS_ERR_BADREAL:		s = "bad real value"; break;
	case LAPIS_ERR_LOAD:		s = "wrong # of data for loading"; break;
	case LAPIS_ERR_BADVARMETHOD:	s = "invalid 'variable::method' expression"; break;
	case LAPIS_ERR_VARIABLE:	s = "unknown variable"; break;
	case 0:
		break;
	default:
		s = "other error";
		printf ("#%d\n", err);
	}

	if (!s) {
		if (stack)
			fprintf (stderr, "Lapis exiting without errors but stack not empty.\n");
		else
			fprintf (stderr, "Lapis thanks you for rock-solid run.\n");
	} else {
		if (word)
			fprintf (stderr, "Lapis error at word '%s': %s\n", word,s);
		else
			fprintf (stderr, "Lapis error: %s\n", s);
	}

	if (stack) {
		fprintf (stderr, "Stack dump:\n");
		int i=0;
		while (stack) {
			StackItem *next = stack->next;
			char *ts = "object ";

			fprintf (stderr, "%d: ", i++);

			switch (stack->type) {
			case STACKITEM_INTEGER: ts = "integer "; break;
			case STACKITEM_BOOLEAN: ts = "boolean "; break;
			case STACKITEM_REAL: ts = "real "; break;
			case STACKITEM_STRING: ts = "string "; break;
			case STACKITEM_STYLE: ts = "style "; break;
			case STACKITEM_SYMBOL: ts = "symbol "; break;
			case STACKITEM_IMAGE: ts = "image "; break;
			case STACKITEM_DICTIONARY: ts = "dictionary "; break;
			case STACKITEM_MATRIX: ts = "matrix "; break;
			case STACKITEM_ARRAY: ts = "array "; break;
			case STACKITEM_PARAMSTART: ts = "param-start "; break;
			case STACKITEM_PARAMEND: ts = "param-end "; break;
			case STACKITEM_TABLE: ts = "table "; break;
			}
			fprintf (stderr, "%s", ts);

			switch (stack->type) {
			case STACKITEM_INTEGER:
			case STACKITEM_BOOLEAN:
			case STACKITEM_REAL:
				StackItem_print (stack, stderr);
				break;
			case STACKITEM_STRING:
			case STACKITEM_STYLE:
			case STACKITEM_SYMBOL:
				fprintf (stderr, "\"");
				StackItem_print (stack, stderr);
				fprintf (stderr, "\"");
				break;
			case STACKITEM_IMAGE: 
			case STACKITEM_DICTIONARY: 
			case STACKITEM_MATRIX: 
			case STACKITEM_ARRAY: 
			case STACKITEM_PARAMSTART: 
			case STACKITEM_PARAMEND: 
			case STACKITEM_TABLE: 
				break;
			default:
				fprintf (stderr, "type=%d", stack->type);
			}
			fprintf (stderr, "\n");
			stack->next=NULL;
			StackItem_delete (stack);
			stack = next;
		}
	}

	fclose(f);
	fclose(inf);
	if (ouf)
		fclose(ouf);
	fclose(hef);
	return 0;
}

void add_method (Class *class, Method *method)
{
	if (!class || !method)
		return;

	char ch = method->name[0];
	int index = isalpha(ch) ? tolower(ch)-'a' : 26;
	method->next = class->methods[index];
	class->methods[index] = method;
}

void add_class (Class *class)
{
	if (!class)
		return;

	printf ("Adding class %s\n", class->name);

	char ch = class->name[0];
	int index = isalpha(ch) ? tolower(ch)-'a' : 26;
	class->next = classes[index];
	classes[index] = class;
}

Class *lookup_class (char *name)
{
	if (!name)
		return NULL;

	char ch = *name;
	int index = isalpha(ch) ? tolower(ch)-'a' : 26;
	Class *class = classes[index];
	while (class) {
		if (!strcmp(class->name,name))
			return class;
		class = class->next;
	}
	return NULL;
}

void init_lapis ()
{
	Class *image = NULL;
	Class *integer = NULL;
	Class *real = NULL;
	Class *string = NULL;
	Class *font = NULL;
	Class *array = NULL;
	Class *dictionary = NULL;
	Class *matrix = NULL;
	Class *table = NULL;
	Class *boolean = NULL;
	Class *null = NULL;

	printf ("This is Lapis, version %s\n", LAPIS_VERSION);

	add_class (string = Class_new ("String"));
	add_method (string, Method_new("new", method_String_new, 0));
	add_method (string, Method_new2("char", method_String, 1, CMD_STRINGCHAR));
	add_method (string, Method_new2("left", method_String, 1, CMD_STRINGLEFT));
	add_method (string, Method_new2("right", method_String, 1, CMD_STRINGRIGHT));
	add_method (string, Method_new2("trunc", method_String, 1, CMD_STRINGTRUNC));
	add_method (string, Method_new("sub", method_String_sub, 2));
	add_method (string, Method_new("reverse", method_String_reverse, 0));
	add_method (string, Method_new("write", method_write, 0));
	add_method (string, Method_new(".", method_print, 0));
	add_method (string, Method_new2("+", method_binary, 0, CMD_ADD));

	add_class (real = Class_new ("Real"));
	add_method (real, Method_new(".", method_print, 0));
	add_method (real, Method_new("sqrt", method_sqrt, 0));
	add_method (real, Method_new2("pow", method_binary, 0, CMD_POW));
	add_method (real, Method_new2("+", method_binary, 0, CMD_ADD));
	add_method (real, Method_new2("-", method_binary, 0, CMD_SUB));
	add_method (real, Method_new2("/", method_binary, 0, CMD_DIV));
	add_method (real, Method_new2("*", method_binary, 0, CMD_MULT));
	add_method (real, Method_new2("&", method_binary, 0, CMD_AND));
	add_method (real, Method_new2("|", method_binary, 0, CMD_OR));
	add_method (real, Method_new2("^", method_binary, 0, CMD_XOR));
	add_method (real, Method_new2("&&", method_binary, 0, CMD_BOOLAND));
	add_method (real, Method_new2("||", method_binary, 0, CMD_BOOLOR));
	add_method (real, Method_new2("^^", method_binary, 0, CMD_BOOLXOR));
	add_method (real, Method_new2("==", method_binary, 0, CMD_EQUAL));
	add_method (real, Method_new2("!=", method_binary, 0, CMD_NOTEQUAL));
	add_method (real, Method_new2("<", method_binary, 0, CMD_LT));
	add_method (real, Method_new2(">", method_binary, 0, CMD_GT));
	add_method (real, Method_new2(">=", method_binary, 0, CMD_GE));
	add_method (real, Method_new2("<=", method_binary, 0, CMD_LE));
	add_method (real, Method_new2("<<", method_binary, 0, CMD_LSHIFT));
	add_method (real, Method_new2(">>", method_binary, 0, CMD_RSHIFT));
	add_method (real, Method_new2("++", method_unary, 0, CMD_INCREMENT));
	add_method (real, Method_new2("--", method_unary, 0, CMD_DECREMENT));
	add_method (real, Method_new2("neg", method_unary, 0, CMD_NEG));
	add_method (real, Method_new2("ln", method_unary, 0, CMD_LN));
	add_method (real, Method_new2("asin", method_unary, 0, CMD_ASIN));
	add_method (real, Method_new2("acos", method_unary, 0, CMD_ACOS));
	add_method (real, Method_new2("atan", method_unary, 0, CMD_ATAN));
	add_method (real, Method_new2("sin", method_unary, 0, CMD_SIN));
	add_method (real, Method_new2("cos", method_unary, 0, CMD_COS));
	add_method (real, Method_new2("tan", method_unary, 0, CMD_TAN));
	add_method (real, Method_new2("sqrt", method_unary, 0, CMD_SQRT));
	add_method (real, Method_new2("cbrt", method_unary, 0, CMD_CBRT));

	add_class (integer = Class_new ("Integer"));
	add_method (integer, Method_new(".", method_print, 0));
	add_method (integer, Method_new("sqrt", method_sqrt, 0));
	add_method (integer, Method_new("write", method_write, 0));
	add_method (integer, Method_new("++", method_unary, 0));
	add_method (integer, Method_new("--", method_unary, 0));
	add_method (integer, Method_new("~", method_unary, 0));
	add_method (integer, Method_new("!", method_unary, 0));
	add_method (integer, Method_new2("pow", method_binary, 0, CMD_POW));
	add_method (integer, Method_new2("+", method_binary, 0, CMD_ADD));
	add_method (integer, Method_new2("-", method_binary, 0, CMD_SUB));
	add_method (integer, Method_new2("/", method_binary, 0, CMD_DIV));
	add_method (integer, Method_new2("%", method_binary, 0, CMD_REMAINDER));
	add_method (integer, Method_new2("*", method_binary, 0, CMD_MULT));
	add_method (integer, Method_new2("&", method_binary, 0, CMD_AND));
	add_method (integer, Method_new2("|", method_binary, 0, CMD_OR));
	add_method (integer, Method_new2("^", method_binary, 0, CMD_XOR));
	add_method (integer, Method_new2("==", method_binary, 0, CMD_EQUAL));
	add_method (integer, Method_new2("!=", method_binary, 0, CMD_NOTEQUAL));
	add_method (integer, Method_new2("<", method_binary, 0, CMD_LT));
	add_method (integer, Method_new2(">", method_binary, 0, CMD_GT));
	add_method (integer, Method_new2(">=", method_binary, 0, CMD_GE));
	add_method (integer, Method_new2("<=", method_binary, 0, CMD_LE));
	add_method (integer, Method_new2("<<", method_binary, 0, CMD_LSHIFT));
	add_method (integer, Method_new2(">>", method_binary, 0, CMD_RSHIFT));
	add_method (integer, Method_new2("&&", method_binary, 0, CMD_BOOLAND));
	add_method (integer, Method_new2("||", method_binary, 0, CMD_BOOLOR));
	add_method (integer, Method_new2("^^", method_binary, 0, CMD_BOOLXOR));
	add_method (integer, Method_new2("++", method_unary, 0, CMD_INCREMENT));
	add_method (integer, Method_new2("--", method_unary, 0, CMD_DECREMENT));
	add_method (integer, Method_new2("~", method_unary, 0, CMD_INVERT));
	add_method (integer, Method_new2("neg", method_unary, 0, CMD_NEG));
	add_method (integer, Method_new2("ln", method_unary, 0, CMD_LN));
	add_method (integer, Method_new2("asin", method_unary, 0, CMD_ASIN));
	add_method (integer, Method_new2("acos", method_unary, 0, CMD_ACOS));
	add_method (integer, Method_new2("atan", method_unary, 0, CMD_ATAN));
	add_method (integer, Method_new2("sin", method_unary, 0, CMD_SIN));
	add_method (integer, Method_new2("cos", method_unary, 0, CMD_COS));
	add_method (integer, Method_new2("tan", method_unary, 0, CMD_TAN));
	add_method (integer, Method_new2("sqrt", method_unary, 0, CMD_SQRT));
	add_method (integer, Method_new2("cbrt", method_unary, 0, CMD_CBRT));

	add_class (image = Class_new ("Image"));
	add_method (image, Method_new("write", method_write, 0));
	add_method (image, Method_new(".", method_print, 0));
	add_method (image, Method_new("new", method_Image_new, 2));
	add_method (image, Method_new("clear", method_Image_clear, 0));
	add_method (image, Method_new2("moveto",method_Image_graphic, 2, CMD_MOVETO));
	add_method (image, Method_new2("lineto", method_Image_graphic, 2, CMD_LINETO));
	add_method (image, Method_new2("fillrect", method_Image_graphic, 2, CMD_FILLRECT));
	add_method (image, Method_new("setcolor", method_Image_setcolor, 1));
	add_method (image, Method_new("setfont", method_Image_setfont, 1));
	add_method (image, Method_new("text", method_Image_show, 1));

	add_class (boolean = Class_new ("Boolean"));
	add_method (boolean, Method_new (".", method_print, 0));
	add_method (boolean, Method_new2("!", method_unary, 0, CMD_NOT));
	add_method (boolean, Method_new2("&", method_binary, 0, CMD_AND));
	add_method (boolean, Method_new2("|", method_binary, 0, CMD_OR));
	add_method (boolean, Method_new2("^", method_binary, 0, CMD_XOR));
	add_method (boolean, Method_new2("&&", method_binary, 0, CMD_BOOLAND));
	add_method (boolean, Method_new2("||", method_binary, 0, CMD_BOOLOR));
	add_method (boolean, Method_new2("^^", method_binary, 0, CMD_BOOLXOR));
	add_method (boolean, Method_new2("==", method_binary, 0, CMD_EQUAL));
	add_method (boolean, Method_new2("!=", method_binary, 0, CMD_NOTEQUAL));
	add_method (boolean, Method_new2("+", method_binary, 0, CMD_ADD));
	add_method (boolean, Method_new2("-", method_binary, 0, CMD_SUB));
	add_method (boolean, Method_new2("*", method_binary, 0, CMD_MULT));
	add_method (boolean, Method_new2("/", method_binary, 0, CMD_DIV));

	add_class (null = Class_new ("Null"));
	add_method (null, Method_new (".", method_print, 0));

	add_class (font = Class_new ("Font"));
	add_method (font, Method_new("write", method_write, 0));
	add_method (font, Method_new(".", method_print, 0));
	add_method (font, Method_new("new", method_Font_new, 0));
	add_method (font, Method_new("getconsolefont", method_Font_getconsolefont, 1));

	add_class (array = Class_new ("Array"));
	add_method (array, Method_new("write", method_write, 0));
	add_method (array, Method_new(".", method_print, 0));
	add_method (array, Method_new("new", method_Array_new, 1));
	add_method (array, Method_new("@", method_Array_get, 1));
	add_method (array, Method_new("!", method_Array_put, 1));
	add_method (array, Method_new(":=", method_Array_load, -1));
	add_method (array, Method_new("size", method_Array_size, 0));

	add_class (dictionary = Class_new ("Dictionary"));
	add_method (dictionary, Method_new("new", method_Dictionary_new, 0));
	add_method (dictionary, Method_new("write", method_write, 0));
	add_method (dictionary, Method_new(".", method_print, 0));
	add_method (dictionary, Method_new("@", method_Dictionary_get, 1));
	add_method (dictionary, Method_new("!", method_Dictionary_put, 1));
	add_method (dictionary, Method_new("-", method_Dictionary_remove, 1));
	add_method (dictionary, Method_new("+", method_Dictionary_combine, 0));
	add_method (dictionary, Method_new("count", method_Dictionary_count, 0));

	add_class (matrix = Class_new ("Matrix"));
	add_method (matrix, Method_new("write", method_write, 0));
	add_method (matrix, Method_new(".", method_print, 0));
	add_method (matrix, Method_new("new", method_Matrix_new, 2));
	add_method (matrix, Method_new("!", method_Matrix_put, 1));
	add_method (matrix, Method_new(":=", method_Matrix_load, -1));
	//add_method (matrix, Method_new("@", method_Matrix_get, 1));

	add_class (table = Class_new ("Table"));
	add_method (table, Method_new(".", method_print, 1));
	// add_method (table, Method_new("write", method_write, 0));
	// add_method (table, Method_new("new", method_Table_new, 0));
	// add_method (table, Method_new("column", method_Table_new, 0));
	// add_method (table, Method_new("select", method_Table_new, 0));
	// add_method (table, Method_new("row", method_Table_new, 0));
}

void myhandler(int foo)
{
	fprintf (stderr, "Error: segmentation fault.\n");
	exit(0);
}

Class *StackItem_getclass (StackItem *im)
{
	if (!im)
		return NULL;

	Class *class = NULL;

	switch (im->type) {
	case STACKITEM_STRING:
		class = lookup_class("String");
		break;
	case STACKITEM_IMAGE:
		class = lookup_class("Image");
		break;
	case STACKITEM_INTEGER:
		class = lookup_class("Integer");
		break;
	case STACKITEM_ARRAY:
		class = lookup_class("Array");
		break;
	case STACKITEM_MATRIX:
		class = lookup_class("Matrix");
		break;
	case STACKITEM_DICTIONARY:
		class = lookup_class("Dictionary");
		break;
	case STACKITEM_REAL:
		class = lookup_class("Real");
		break;
	case STACKITEM_BOOLEAN:
		class = lookup_class("Boolean");
		break;
	case STACKITEM_PARAMSTART:
	case STACKITEM_PARAMEND:
		class = NULL;
		break;
	case STACKITEM_NULL:
		class = lookup_class("Null");
		break;
	case STACKITEM_TABLE:
		class = lookup_class("Table");
		break;
	}
	return class;
}

// Process any Lapis program
//
bool process_lapis (int fd, char *localpath)
{
	char inputpath [30];
	char outputpath [30];
	char headerpath [30];

	signal (SIGSEGV, myhandler);

	maketempfile (inputpath);
#ifndef TESTING
	maketempfile (outputpath);
#else
	strcpy (outputpath, "out");
#endif
	maketempfile (headerpath);

	global_vars = Dictionary_new();

	FILE *f = fopen (inputpath, "w");
	fprintf (f, "foo=1\n");
	fclose (f);

	process_lapis_core (fd, localpath, inputpath, outputpath, headerpath);

	unlink (inputpath); // POST params

#ifndef TESTING
	if (!send_results (fd, outputpath, headerpath)) {
		unlink (outputpath);
		unlink (headerpath);
		log_write ("\tERROR: Lapis send failed\n");
		http_error (fd, HTTP_ERR_INTERNAL);
	}
#endif
	
	unlink (headerpath);
#ifndef TESTING
	unlink (outputpath);
	log_store ();
#endif
	return myexit(0);
}

#endif
