
/*===========================================================================
 * Lapis, a simple webserver.
 * Copyright (C) 2005,2011 by Zachary Smith.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 1, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *=========================================================================*/

/*---------------------------------------------------------------------------
 * Change log
 *
 * 2005 Jul 01: conception
 * 2005 Jul 02: added automatic parsing of http header
 * 2005 Jul 02: added logging of http header
 * 2005 Jul 04: added very basic PHP support (no cookies, no URL params)
 * 2005 Jul 10: added witty retort
 * 2005 Jul 19: improved log reporting.
 * 2005 Jul 20: added support for absolute path location for HTML etc.
 * 2005 Jul 24: added support for Perl
 * 2005 Jul 24: made PHP support optional
 * 2005 Jul 24: fixed zombie problem.
	=> 0.4
 * 2005 Jul 25: genericized preload mechanism, created preload_file etc.
	=> 0.5
 * 2005 Jul 25: added IP blocking for when hosts.deny is unreliable.
	=> 0.6
 * 2005 Jul 26: improved logging with file lock.
	=> 0.7
 * 2005 Jul 29: added log buffer, to reduce delays waiting for log file.
 * 2005 Aug 01: added automatic rejection of requests with "Authorization".
 * 2005 Aug 01: program now remembers specific attacker IPs and blocks them.
	=> 0.8
 * 2005 Aug 01: added blocking when too few params (typical of attackers).
 * 2005 Aug 01: parameter parser now produces parameter numbers.
 * 2005 Aug 01: added limit of only 1 active request at a time per IP,
 *		with one more queueable.
 * 2005 Aug 03: added load/save of attacker list to/from file.
	=> 0.9
 * 2005 Aug 04: added optional Python support.
 * 2005 Aug 04: made Perl support optional.
 * 2005 Aug 04: added proper file extension parsing.
 * 2005 Aug 05: expanded number of waiting requests to unlimited.
 * 2005 Aug 06: added checking for and validation of User-agent.
	=> 0.10
 * 2005 Aug 06: added ability to run stack language (Lapis stack language)
 * 2005 Aug 06: conception of simple stack language interpreter
 * 2005 Aug 06: added arithmetic and logic ops to stack language
 * 2005 Aug 06: added basic text attributes to stack language
 * 2005 Aug 06: added table creation functions to stack language
 * 2005 Aug 06: added style and DIV creation functions to stack language
	=> 0.11
 * 2005 Aug 08: added set-cookie command to stack language
 * 2005 Aug 08: added HTTP_COOKIE support
 * 2005 Aug 08: fixed tempfile not getting unlinked bug.
 * 2005 Aug 08: separated stack language code from Lapis code.
 * 2005 Aug 16: added support for TIFF
 * 2005 Aug 19: fixed upper-case extension rejection bug.
	=> 0.12
 * 2005 Aug 20: Stack language optimizations.
	=> 0.13
 * 2005 Aug 20: added detection of (one) repeat zero-byte attacker.
 * 2005 Aug 20: added detection of >1 content-length => poisoning attack.
	=> 0.14
 *-------------------------------------------------------------------------*/


#include <stdio.h>
#include <netdb.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>	// socket, bind ec
#include <sys/file.h>	// flock
#include <unistd.h>
#include <stdlib.h>
#include <time.h>
#include <ctype.h>	// tolower
#include <errno.h>
#include <signal.h>	// sigaction
#include <sys/wait.h>	// waitpid
#include <fcntl.h>	// file lock

#include "defs.h"
#include "lapis.h"


bool process_script (int fd, char *localpath, char type);


static int webdirlen;

static IP last_zerobyte_ip = 0;
static char zerobyte_count = 0;

static char header[2000];
static int headerlen = 0;

static unsigned char ip_block_all [256];
static short ip_block_sub1 [256];
static short ip_block_sub2 [256];

#define MAXATTACKERS 1024
static IP	*attackers;
static int total_attackers = 0;

#define IPBLOCK_ALL(num) { ip_block_all[num]=1; }
#define IPBLOCK_SUB1(high,num) { ip_block_sub1[high]=num; }
#define IPBLOCK_SUB2(high,num) { ip_block_sub1[high]=num; }

#define MAXHASH 256
#define HASHAND 255
static ListItem *pid_hash[MAXHASH];

static AddrListItem *addr_hash[256]; // key is last # of IP

#define MAXLOG 65536
static char my_log [MAXLOG];
static int my_log_ix = 0;


static char *latin_text = "Quod ad domum Milonis Hypatini quam proxime diripuimus pertinet, discussa sollicitudine iam possumus esse securi. Postquam vos enim fortissimis viribus cunctus ablatis castra nostra remeastis, immixtus ego turbelis popularium dolentique atque indignanti similis arbitrabar super investigatione facti cuius modi consilium caperetur et an et quatenus latrones placeret inquiri, renuntiaturus vobis, uti mandaveratis, omnia. nec argumentis dubiis, sed rationibus probabilibus congruo cunctae multitudinis consensu nescio qui Lucios auctor manifestus facinoris postulabatur, qui proximis diebus fictis commendaticiis litteris Miloni sese virum commentitus bonum artius conciliaverat, ut etiam hospitio susceptus inter familiaris intimos haberetur, plusculisque ibidem diebus demoratus falsis amoribus ancillae Milonis animum inrepens ianuae claustra sedulo exploraverat et ipsa membra in quis omne patrimonium condi solebat curiose perspexerat.";

static int latin_text_len;

typedef struct param {
	char *name;
	char code;
} KnownParam;

static KnownParam known_params [] = {
	{ "Host", PARAM_HOST },
	{ "Referer", PARAM_REFERER },
	{ "User-Agent", PARAM_USERAGENT},
	{ "Accept", PARAM_ACCEPT },
	{ "Accept-Language", PARAM_ACCEPTLANGUAGE },
	{ "Accept-Encoding", PARAM_ACCEPTENCODING },
	{ "Accept-Charset", PARAM_ACCEPTCHARSET },
	{ "Keep-Alive", PARAM_KEEPALIVE },
	{ "Connection", PARAM_CONNECTION },
	{ "Authorization", PARAM_AUTHORIZATION },
	{ "Cache-control", PARAM_CACHECONTROL },
	{ "Cookie", PARAM_COOKIE },
	{ "Content-length", PARAM_CONTENTLENGTH},
	{ NULL, 0 }
};


void log_store ()
{
	FILE *f;
	while (!(f = fopen ("/tmp/lapis.log", "a")))
		;
	int i = flock (fileno (f), LOCK_EX);
	if (i) {
		perror ("flock");
		myexit(EXITCODE_FLOCK_FAILED); // flock should block
	}
	fwrite (my_log, 1, my_log_ix, f);
	fclose (f);

	my_log_ix = 0;
}


void log_write (char *s)
{
	if (!s) 
		return;

	int len = strlen(s);
	int remaining = MAXLOG - my_log_ix - 10;

	if (len > remaining) {
		strncpy (&my_log[my_log_ix], s, remaining);
		my_log_ix += remaining;
		my_log[my_log_ix] = 0;
	} else {
		strcpy (&my_log[my_log_ix], s);
		my_log_ix += len;
	}
}

void log_writeln_num (int n)
{
	char tmp[20];
	sprintf (tmp, "%2d\n", n);
	log_write (tmp);
}

void log_write_num (int n)
{
	char tmp[20];
	sprintf (tmp, "%2d", n);
	log_write (tmp);
}

void log_writeln (char *s)
{
	if (s)
		log_write (s);
	log_write ("\n");
}


char parse_extension (char *s)
{
	if (!s)
		return EXTN_NONE;

	char c= tolower(*s++);
	char c2,c3;
	switch (c) {
	case '-':
		c2 = *s++;
		if (!c2)
			return EXTN_LAPIS;
		break;
	case 'i':
		if (!strcasecmp ("co",s))
			return EXTN_ICO;
		break;
	case 'h':
		if (!strcasecmp ("tml",s) || !strcasecmp ("tm",s))
			return EXTN_HTML;
		break;
	case 'j':
		if (!strcasecmp ("pg",s) || !strcasecmp ("peg",s))
			return EXTN_JPG;
		c2 = tolower(*s++);
		if (c2=='s' && !*s)
			return EXTN_JS;
		break;
	case 't':
		if (!strcasecmp ("if",s))
			return EXTN_TIF;
		if (!strcasecmp ("iff",s))
			return EXTN_TIF;
		break;
	case 'g':
		if (!strcasecmp ("if",s))
			return EXTN_GIF;
		break;
	case 'o':
		if (!strcasecmp ("gg",s))
			return EXTN_OGG;
		break;
	case 'a':
		if (!strcasecmp ("vi",s))
			return EXTN_AVI;
		break;
	case 'm':
		c2 = tolower(*s++);
		if (c2 == 'p') {
			char c3 = tolower(*s++);
			if (c3 == 'g' && !*s)
				return EXTN_MPG;
			else if (c3 == '3' && !*s)
				return EXTN_MP3;
		}
		break;
	case 'p':
		c2 = tolower(*s++);
		switch (c2) {
		case 'l':
			if (!*s)
				return EXTN_PL;
			break;
		case 'y':
			if (!*s)
				return EXTN_PY;
			break;
		case 'h': 
			c3=tolower(*s++);
			if (c3=='p' && !*s)
				return EXTN_PHP;
			break;
		case 'n':
			c3=tolower(*s++);
			if (c3=='g' && !*s)
				return EXTN_PNG;
			break;
		}
		break;
	}
	return EXTN_NONE;
}

char lookup_param (char *s)
{
	int i=0;
	char *s2;
	if (!s)
		return 0;
	while ((s2 = known_params[i].name)) {
		if (!strcasecmp (s2, s))
			return known_params[i].code;
		i++;
	}
	return -1;
}

void ip_to_str(IP addr, char *str)
{
	if (!str)
		return;

	unsigned short a = 0xff & addr;
	unsigned short b = 0xff & (addr >> 8);
	unsigned short c = 0xff & (addr >> 16);
	unsigned short d = 0xff & (addr >> 24);
	sprintf (str, "%d.%d.%d.%d", a,b,c,d);
}

	
short get_pid_hash_key (pid_t pid)
{
	if (pid < 0)
		return -1;

	return HASHAND & (unsigned)pid;
}

ListItem *pid_hash_remove (pid_t pid)
{
	if (pid < 0)
		return NULL;

	short key = get_pid_hash_key (pid);
	if (key < 0)
		return NULL;

	ListItem *im = pid_hash[key];
	ListItem *prev = NULL;
	bool found = 0;

	while (im) {
		if (im->pid == pid) {
			found = 1;
			break;
		}

		prev = im;
		im = im->next;
	}

	if (!found)
		return NULL;

	if (!prev) 
		pid_hash[key] = im->next;
	else 
		prev->next = im->next;

	return im;
}

AddrListItem *addr_hash_lookup (IP addr)
{
	short key = addr >> 24;

	AddrListItem *im = addr_hash[key];

	while (im) {
		if (im->addr == addr) {
			return im;
		}

		im = im->next;
	}
	return NULL;
}

AddrListItem *addr_hash_add (IP addr)
{
	unsigned short key = addr >> 24;

	AddrListItem *existing;
	existing = addr_hash_lookup (addr);
	if (existing)
		return existing;

	AddrListItem *nu = (AddrListItem*) malloc (sizeof (AddrListItem));
	if (!nu) {
		fprintf (stderr, "Out of memory\n");
		log_store ();
		myexit(EXITCODE_NOMEM);
	}

	nu->list = NULL;
	nu->addr = addr;
	nu->next = addr_hash[key];
	addr_hash[key] = nu;
	return nu;
}


// If delayed processing, set pid = -1
// To update an already existing entry, just set the pid.
//
bool addr_hash_add_fd (IP addr, int fd, int pid)
{
	AddrListItem *im = addr_hash_lookup (addr);
	if (!im) {
		im = addr_hash_add (addr);
		if (!im) {
			return false;
		}
	}

	AddrListSubItem *sim;
	char tmp[30];
	ip_to_str (addr,tmp);

	sim = malloc (sizeof(AddrListSubItem));
	if (!sim)
		return false;

	sim->fd = fd;
	sim->pid = pid;
	sim->next = im->list;
	im->list = sim;
	return true;
}


// Find out how many requests we are currently processing 
// for a given IP.
int addr_hash_count_active (IP addr)
{
	int count_active=0;
	int count_inactive=0;

	AddrListItem *im = addr_hash_lookup (addr);
	if (!im) {
		return -1;
	}

	AddrListSubItem *sim = im->list;
	if (!sim) {
		warning ("AddrListItem exists without list");
	}

	while (sim) {
		if (sim->pid != -1) {
			count_active++;
		}
		else
			count_inactive++;
		sim = sim->next;
	}

#if 0
char tmp[30];
ip_to_str (addr,tmp);
printf ("found that IP %s has %d active connections, %d inactive\n",
	tmp,count_active,count_inactive);
#endif

	return count_active;
}


	
// Remove all of the entries for a given IP.
//
int addr_hash_remove_all (IP addr)
{
	unsigned short key = addr >> 24;

	AddrListItem *im = addr_hash[key];
	AddrListItem *prev = NULL;

	while (im) {
		if (im->addr == addr) {
			break;
		}

		prev = im;
		im = im->next;
	}

	if (!im) {
		return -1;
	}

	if (!prev)
		addr_hash[key] = im->next;
	else 
		prev->next = im->next;

	// Remove all the connections.
	AddrListSubItem *sim = im->list;
	int count=0;
	while (sim) {
		count++;
		AddrListSubItem *next = sim->next;
		close (sim->fd);
		free (sim);
		sim = next;
	}
	im->list = NULL;
	free (im);

	return count;
}

	
AddrListSubItem *addr_hash_remove_one (IP addr, int pid)
{
	AddrListItem *im = addr_hash_lookup (addr);
	if (!im) {
		return NULL;
	}

	AddrListSubItem *prev = NULL;
	AddrListSubItem *sim = im->list;
	if (!sim) {
		return NULL;
	}
	while (sim && pid != sim->pid) {
		prev = sim;
		sim = sim->next;
	}
	if (!sim) {
		return NULL;
	}
	if (prev) {
		prev->next = sim->next;
	} else {
		im->list = sim->next;
	}
	sim->next = NULL;

#if 0
char tmp[30];
ip_to_str (addr,tmp);
printf ("removed IP %s, pid %d\n", tmp,sim->pid);
#endif

	if (!im->list)
		addr_hash_remove_all (addr);

	return sim;
}

void pid_hash_add (pid_t pid, IP addr)
{
	if (pid < 0)
		return;

	short key = get_pid_hash_key (pid);
	if (key < 0)
		return;

	ListItem *tmp;
	if ((tmp = pid_hash_remove (pid))) {
		fprintf (stderr, "Process id already in pid hash!\n");
	}

	ListItem *nu = (ListItem*) malloc (sizeof (ListItem));
	if (!nu) {
		fprintf (stderr, "Out of memory\n");
		log_store ();
		myexit(EXITCODE_NOMEM);
	}

	nu->pid = pid;
	nu->addr = addr;
	nu->next = pid_hash[key];
	pid_hash[key] = nu;
}



int process_request (int fd, struct sockaddr_in *from);


/* In-memory cacheing of specific files.
 */
#define MAXPRELOAD 8
static unsigned char *preload [MAXPRELOAD];
static int preload_len [MAXPRELOAD];
enum {
	PRELOAD_INDEX = 0,
	PRELOAD_FAVICON = 1
};


bool load_attackers ()
{
	FILE *f = fopen (ATTACKERS_FILE, "r");
	if (!f) {
		log_writeln ("Unable to read attackers list.");
		log_store ();
		return false;
	}

	int count=0;
	if (1 != fscanf(f, "%d", &count)) {
		log_writeln ("Bad format in attackers list.");
		log_store ();
		fclose(f);
		return false;
	}

	if (count > MAXATTACKERS)
		count = MAXATTACKERS;
	total_attackers = count;

	int i;
	for (i=0; i<count; i++)
		attackers[i]=0;

	for (i=0; i<count; i++) {
		unsigned long a,b,c,d;
		if (4 != fscanf (f, "%lu.%lu.%lu.%lu", &a,&b,&c,&d)) {
			log_writeln ("Bad format in attackers list.");
			log_store ();
			fclose(f);
			total_attackers = count;
			return false;
		}
		a &= 0xff;
		b &= 0xff;
		c &= 0xff;
		d &= 0xff;
		b <<= 8;
		c <<= 16;
		d <<= 24;
		attackers[i] = a|b|c|d;
	}

	log_writeln (NULL);
	log_write ("Total attacker IPs read from file: ");
	log_writeln_num (total_attackers);
	log_writeln (NULL);
	log_store ();

	return true;
}


bool save_attackers ()
{
	FILE *f = fopen (ATTACKERS_FILE, "w");
	if (!f) {
		log_writeln ("Unable to write attackers list.");
		log_store ();
		return false;
	}

	char ipstr[30];
	int i;
	fprintf (f, "%d\n", total_attackers);
	for (i=0; i<total_attackers; i++) {
		ip_to_str (attackers[i], ipstr);
		fprintf (f, "%s\n", ipstr);
	}
	fclose (f);
	return true;
}

void add_attacker (IP addr)
{
	if (total_attackers >= MAXATTACKERS) {
		log_write ( "too many attackers\n");
		log_store ();
		return;
	}

	attackers [total_attackers++] = addr;
	save_attackers();
}


unsigned char *get_file (char *path, int *lenreturn, bool prepend)
{
	struct stat st;
	char fullpath [MAXPATH];

	if (!path)
		return NULL;
	/*----------*/
printf ("get_file %s\n", path);
	if (prepend) {
		strcpy (fullpath, WEBDIR);	// must include trailing '/'
		strncpy (fullpath+webdirlen, path, MAXPATH-2-webdirlen);
	} else 
		strncpy (fullpath, path, MAXPATH-2);

	fullpath[MAXPATH-1] = 0;

	// Get file length.
	if (stat (fullpath, &st)) {
		perror ("stat");
		return NULL;
	}
	int file_len = st.st_size;

	// Allocate buffer.
	unsigned char *data = (char*) malloc (file_len+1);
	if (!data) {
		perror ("malloc");
		log_writeln ("out of memory");
		log_store ();
		myexit(EXITCODE_NOMEM);
	}

	// Read the entire file in one go.
	FILE *f = fopen (fullpath, "rb");
	if (!f)
		return NULL;
	int i = fread (data, 1, file_len, f);
	fclose (f);
	if (i != file_len)  {
		free (data);
		return NULL;
	}

	// Done. Return length.
	if (lenreturn)
		*lenreturn = file_len;

	// Store an end-of-string just in case.
	data[file_len] = 0;

	// Return file data.
	return data;
}

int preload_file (char *path, int which)
{
	int len;
	char *s;
	if (which < 0 || which >= MAXPRELOAD)
		return false;
	if (preload[which])
		return false;
	s = preload[which] = get_file (path, &len, 1);
	if (!s) {
		return false;
	}
	preload_len [which] = len;
	return true;
}


int parse_http_header (char *s, StringPair results[], int max_results,
			int *referer_index_return, int *flags_return,
			int *connection_return)
{
	int count=0;
	int total_contentlength=0;

	if (!s) 
		return 0;
	if (max_results < 1) 
		return 0;
	if (!flags_return)
		return 0;
	/*---------*/

	if (flags_return)
		*flags_return = 0;
	if (connection_return)
		*connection_return = 0;

	// Skip the get/post
	char *foo = strchr (s, '\n');
	if (!foo)
		foo = strchr (s, '\r');
	if (!foo)
		return 0;
	s = foo+1;
	if (*s == '\n' || *s == '\r')
		s++;

	// Keep track of User-agent
	char found_useragent=false;

	// Process the data
	while (*s && count < max_results) 
	{
		while (*s && (*s==' ' || *s=='\t' || *s=='\r'))
			s++;

		char *s2 = strchr (s, '\n');
		if (!s2)
			s2 = strchr (s, '\r');
		if (!s2)
			break;
		*s2 = 0;

		char *colon = strchr (s, ':');
		if (!colon)
			break;
		*colon = 0;

		char *contents = colon+1;
		while (*contents && (*contents==' ' || *contents=='\t'))
			contents++;

		char *junk = strchr (contents, '\r');
		if (junk)
			*junk=0;

		char code = lookup_param (s);
		results[count].name = strdup (s);
		results[count].code = code;
		if (code < 0) {
			log_write ("Unknown param: ");
			log_writeln (s);
		}
		if (code == PARAM_CONNECTION && connection_return) {
			if (!strcasecmp ("close", contents))
				*connection_return = 1;
			else
			if (!strcasecmp ("keep-alive", contents))
				*connection_return = 2;
			else
				*connection_return = 3;
		} else
		if (code == PARAM_CONTENTLENGTH) {
			++total_contentlength;
			if (total_contentlength > 1) {
				// cache poisoning attack
				exit (EXITCODE_POISONING);
			}
		} else
		if (code == PARAM_REFERER)
			*referer_index_return = count;
		else
		if (flags_return && code == PARAM_AUTHORIZATION)
			*flags_return |= HEADER_HAS_AUTHORIZATION;
		else
		if (code == PARAM_USERAGENT) {
			if (my_strcasestr (s, "Dillo") ||
			    my_strcasestr (s, "Lynx") ||
			    my_strcasestr (s, "Mozilla") ||
			    my_strcasestr (s, "Explorer") ||
			    my_strcasestr (s, "Netscape"))
				found_useragent = true;
		}
		else
		if (code == PARAM_COOKIE) {
			setenv ("HTTP_COOKIE", contents, 1);
		}

		results[count].contents = strdup (contents);
		count++;

		s = s2+1;
	}

	if (!found_useragent)
		*flags_return |= HEADER_BAD_USERAGENT;

	return count;
}

int 
main (int argc, char **argv)
{
	int i;
	int listensock;
	int clientsock;
	struct servent *sp;
	struct sockaddr_in sin, from;

#ifdef SUPPORT_LAPIS
	init_lapis ();
#endif

	my_log_ix = 0;

	log_writeln(NULL);
	log_writeln ("=============================================================================");
	log_write ("Lapis server started @ ");
	time_t t = time(NULL);
	log_write (ctime(&t)); // ctime appends newline
	log_writeln ("=============================================================================");
	log_store ();

	attackers = (IP*) malloc(MAXATTACKERS * 4);
	if (!attackers) {
		fprintf (stderr, "out of memory");
		return 0;
	}
	load_attackers();

	latin_text_len = strlen (latin_text);

	// Clear the IP blocking array.
	for (i=0; i<256; i++) {
		ip_block_all [i] = 0;
		ip_block_sub1[i] = -1;
		ip_block_sub2[i] = -1;
	}

	for (i=0; i<MAXHASH; i++)
		pid_hash[i] = NULL;

	for (i=0; i<256; i++)
		addr_hash[i] = NULL;

	// Lock out the hackers
	// This is necessary because hosts.deny has proven ineffective.
	IPBLOCK_ALL (201);
	IPBLOCK_ALL (202);
	IPBLOCK_ALL (203);
	IPBLOCK_ALL (204);
	IPBLOCK_ALL (205);
	IPBLOCK_ALL (206);
	IPBLOCK_ALL (207);
	IPBLOCK_ALL (208);
	IPBLOCK_ALL (209);
	IPBLOCK_ALL (210);
	IPBLOCK_ALL (216);
	IPBLOCK_ALL (211);
	IPBLOCK_ALL (212);
	IPBLOCK_ALL (213);
	IPBLOCK_ALL (214);
	IPBLOCK_ALL (215);
	IPBLOCK_ALL (217);
	IPBLOCK_ALL (218);
	IPBLOCK_ALL (219);
	IPBLOCK_ALL (220);
	IPBLOCK_ALL (221);
	IPBLOCK_ALL (222);
#if 1
	IPBLOCK_SUB1(80,17);
	IPBLOCK_SUB1(60,36);
	IPBLOCK_SUB1(195,70);
#endif

	// Set up boilerplate header.
	//
	sprintf (header, "\
HTTP/1.0 200 OK\r\n\
Server: Lapis version %s\n\
Connection: close\n", RELEASE);
	headerlen = strlen (header);

	// Time saver.
	webdirlen = strlen(WEBDIR);

	// Read in the index to improve speed.
	//
	preload_file ("index.html", PRELOAD_INDEX);
	preload_file ("favicon.ico", PRELOAD_FAVICON);

	// Banner
	//
	printf ("\nThis is Lapis webserver, version %s\n", RELEASE);
	printf ("Copyright (C) 2005,2011 by Zachary Smith.\n\n\
This program is distributed in the hope that it will be useful,\n\
but WITHOUT ANY WARRANTY; without even the implied warranty of\n\
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n\
file COPYING, which contains the GNU General Public License,\n\
for more details.\n\n");

#if 0
	char *service = "http";
	if ((sp = getservbyname (service, "tcp")) == NULL) {
		perror ("getservbyname");
		log_store();
		error (1);
	}
	sin.sin_port = sp->s_port;
#else
//#define LISTEN_PORT 47777
#define LISTEN_PORT 80
	sin.sin_port = htons(LISTEN_PORT);
#endif

	sin.sin_family = AF_INET;
	sin.sin_addr.s_addr = htonl(INADDR_ANY);

	if (( listensock = socket (AF_INET, SOCK_STREAM, 0)) < 0)  {
		perror ("socket");
		log_store();
		error (2);
	}

	if (bind (listensock, (struct sockaddr*) &sin, sizeof(sin)) < 0) {
		perror ("bind");
		log_store();
		error (3);
	}

	if (listen (listensock, 500) < 0) {
		perror ("listen");
		log_store();
		error (4);
	}

	//----------
	// Main loop
	//
	while (1) {
		// Wait for a client to contact us.
		socklen_t len;
		clientsock = accept(listensock, 
				(struct sockaddr*) &from, &len);
		if (clientsock < 0) {
			perror ("accept");
			log_store();
			error (5);
		}

		if (len != sizeof (struct sockaddr_in)) {
			log_store();
			error (6);
		}
		
		IP addr = from.sin_addr.s_addr;

		char found_attacker=0;
		i=0;
		while (!found_attacker && i < total_attackers) {
			IP m = attackers[i++];
			if (addr != m)
				continue;

			found_attacker=1;
		}
		if (found_attacker) {
			char ipstring[30];
			ip_to_str (addr, ipstring);
			fprintf (stderr, "Attacker from %s blocked\r\n", ipstring);

			close(clientsock);
			continue;
		}

		// Here we shall refuse to deal with more than one
		// request at a time from any given IP, because one form 
		// of attack involves making many simultaneous requests.
		//
		bool dont_execute = false;
		int count = addr_hash_count_active (addr);
		if (count >= 1) {
			dont_execute = 1;
			addr_hash_add_fd (addr, clientsock, -1);
		}

		AddrListSubItem *pending=NULL;
process_more:	if (pending)
			dont_execute=false;

		// Create a child process to deal with request.
		if (!dont_execute) {
			int pid;

			if (!(pid = fork())) {
				my_log_ix = 0;
				process_request (clientsock, &from);
				myexit (0);
			}

			// Let's keep track of the child
			pid_hash_add (pid, addr);

			if (!pending)
				addr_hash_add_fd (addr, clientsock, pid);
			else
				pending->pid = pid;

			// Release parent copy of socket.
			close (clientsock);
		}

		// Check up on children that have exited
		int child, status;
		while ((child = waitpid (-1, &status, WNOHANG))) {
			if (child < 0) {
perror ("waitpid");
				break;
			}

			if (WIFEXITED (status)) {
				status = WEXITSTATUS (status);
			} else
				status = -1;

			ListItem *im = pid_hash_remove (child);
			if (!im) {
				fprintf (stderr, "Failed to remove child pid %d from hash\n", child);
				continue;
			}

//HERE IS WHERE THE SEGFAULT OCCURRED
			AddrListSubItem *sim = addr_hash_remove_one (im->addr, 
					child);
			if (!sim) {
				fprintf (stderr, "Failed to remove addr %08lx, pid %d from hash\n", im->addr,child);
				continue;
			}
			free(sim);

			char ipstring[30];
			ip_to_str (im->addr, ipstring);

#if 0
			// Debugging stuff!
			FILE *ff = stdout;
			fprintf (ff, "ip=%s\t", ipstring);
			switch (status) {
			case EXITCODE_HTTPERR_MOVED:
				fprintf (ff, "http: moved\n");
				break;
			case EXITCODE_READ_ERROR:
				fprintf (ff, "read error\n");
				break;
			case EXITCODE_HTTPERR_BAD_REQUEST:
				fprintf (ff, "http: bad request\n");
				break;
			case EXITCODE_HTTPERR_UNAUTH:
				fprintf (ff, "http: unauthorized\n");
				break;
			case EXITCODE_HTTPERR_FORBIDDEN:
				fprintf (ff, "http: forbidden\n");
				break;
			case EXITCODE_HTTPERR_NOT_FOUND:
				fprintf (ff, "http: not found\n");
				break;
			case EXITCODE_HTTPERR_INTERNAL:
				fprintf (ff, "http: internal\n");
				break;
			case EXITCODE_HTTPERR_OTHER:
				fprintf (ff, "http: other\n");
				break;
			case EXITCODE_PORTSCAN:
				fprintf (ff, "port scanner\n");
				break;
			case EXITCODE_PARAM_ATTACK:
				fprintf (ff, "parameter attack\n");
				break;
			case EXITCODE_BUFFER_ATTACK:
				fprintf (ff, "buffer attack\n");
				break;
			case EXITCODE_FEW_PARAMS:
				fprintf (ff, "too few parameters\n");
				break;
			case EXITCODE_FLOCK_FAILED:
				fprintf (ff, "flock failed\n");
				break;
			case EXITCODE_NOMEM:
				fprintf (ff, "out of memory\n");
				break;
			case EXITCODE_HTTP_ATTACK:
				fprintf (ff, "http:// atack\n");
				break;
			default:
				fprintf (ff, "other result %d\n", status);
			}
			//fclose(ff);
#endif

			// Two zero byte requests from the same IP
			// in a row mean that we must block that IP.
			//
			if (im->addr == last_zerobyte_ip) {
				if (status == EXITCODE_SHORT) {
					zerobyte_count++;
					if (zerobyte_count==2) {
						status = EXITCODE_PORTSCAN;
						last_zerobyte_ip = 0;
						zerobyte_count = 0;
					}
				} else {
					last_zerobyte_ip = 0;
					zerobyte_count = 0;
				}
			} else {
				if (status == EXITCODE_SHORT) {
					last_zerobyte_ip = im->addr;
					zerobyte_count = 1;
				}
			}

			// Do we have a attacker?
			if (status == EXITCODE_PORTSCAN ||
			    status == EXITCODE_PARAM_ATTACK ||
			    status == EXITCODE_FEW_PARAMS ||
			    status == EXITCODE_HTTP_ATTACK ||
			    status == EXITCODE_BADREQ ||
			    status == EXITCODE_POISONING ||
			    status == EXITCODE_BUFFER_ATTACK ||
			    status == EXITCODE_READ_ERROR) 
			{
				log_write ("Probable attacker at ");
				log_writeln (ipstring);
				add_attacker (im->addr);
				int num = addr_hash_remove_all (im->addr);
				log_write ("Removed ");
				log_write_num (num);
				log_writeln (" pending connections.");
				log_store ();
			} else {
				AddrListItem *im2 = addr_hash_lookup (im->addr);
				if (im2 && im2->list) {
					// Just grab the first one in the list.
					addr = im->addr;
					clientsock = im2->list->fd;
					pending = im2->list;
					//printf ("pending fd=%d\n",clientsock);
					free (im);
					goto process_more;
				} else {
					//printf ("no pending connection\n");
				}
			}

			free (im);
		}
	}

	close (listensock);
	free (attackers);
	return 0;
} // main()


int 
http_error (int fd, int which)
{
	char str[2000];

	log_store ();

	sprintf(str, "HTTP/1.0 %d\r\nConnection: close\r\n\r\n", which);
	write (fd, str, strlen(str));
	write (fd, "\r\n", 2);
	char *s = NULL;
	switch (which) {
	case HTTP_ERR_NOT_FOUND: 
		s = "File not found"; 
		break;
	case HTTP_ERR_MOVED: 
		s = "File has moved"; 
		break;
	case HTTP_ERR_BAD_REQUEST: 
		s = "Bad request"; 
		break;
	case HTTP_ERR_UNAUTH: 
		s = "Unauthorized"; 
		break;
	case HTTP_ERR_FORBIDDEN: 
		s = "Forbidden / Vietato / Verboten / Interdit"; 
		break;
	case HTTP_ERR_INTERNAL: 
		s = "Internal error"; 
		break;
	case HTTP_ERR_BAD_EXTN:
		s = "File extension not supported"; 
		break;
	default: s = "??";
	}

	sprintf (str, "<html>\
<head>\
<body alink=yellow vlink=green link=green bgcolor=#f8f8f8>\n \
<font color=#000080><p>\n\
<table width=100%% border=0>\
<tr>\
<td bgcolor=#e0e0ff>\n\
<font color=black size=+3><b> An error has occurred:\n\
</b><i><small> %s (number %d)\n</small></i></td></tr></table>\n\n", s, which);

	write (fd, str, strlen(str));
	write (fd, latin_text, latin_text_len);

	close (fd);

	switch (which) {
	case HTTP_ERR_NOT_FOUND: 
		myexit(EXITCODE_HTTPERR_NOT_FOUND);
	case HTTP_ERR_MOVED: 
		myexit(EXITCODE_HTTPERR_MOVED);
	case HTTP_ERR_BAD_REQUEST: 
		myexit(EXITCODE_HTTPERR_BAD_REQUEST);
	case HTTP_ERR_UNAUTH: 
		myexit(EXITCODE_HTTPERR_UNAUTH);
	case HTTP_ERR_FORBIDDEN: 
		myexit(EXITCODE_HTTPERR_FORBIDDEN);
	case HTTP_ERR_INTERNAL: 
		myexit(EXITCODE_HTTPERR_INTERNAL);
	default: 
		myexit(EXITCODE_HTTPERR_OTHER);
	}
	return 0;
}


int 
send_feces (int fd)
{
	char *feces = "<html><body><font size=+3>Your IP address looks like that of a hacker, therefore I must repel you. If this is not the case, please try accessing this site from a different computer, or choose a more responsible ISP.";
	int feces_len = strlen (feces);

	char lengthstr[200];
	sprintf (lengthstr, "Content-length: %d\r\n", feces_len);

	write (fd, header, headerlen);
	write (fd, lengthstr, strlen(lengthstr));
	write (fd, "\r\n", 2);
	
	int len = write (fd, feces, feces_len);
	close (fd);

	bool success = len == feces_len;

	log_write ("   \\\n    -> Sent feces");
	if (success)
		log_writeln(NULL);
	else
		log_writeln(" (failed).");

	log_store ();
	return myexit(EXITCODE_HTTP_ATTACK);
}

int send_witty (int fd)
{
	char *witty = "<html><body><font size=+3>Only a girly-man would make a request such as you have just made.</html>";
	int wittylen = strlen(witty);

	char lengthstr[200];
	sprintf (lengthstr, "Content-length: %d\r\n", wittylen);
	write (fd, header, headerlen);
	write (fd, lengthstr, strlen(lengthstr));
	write (fd, "\r\n", 2);
	
	int len;
	len = write (fd, witty, wittylen);
	close (fd);

	bool success = len == wittylen;

	log_write ("   \\\n    -> Sent wittiness");
	if (success)
		log_writeln(NULL);
	else
		log_writeln(" (failed).");

	log_store ();
	return myexit(0);
}

int send_preload (int fd, int which)
{
	char lengthstr[200];

	if (which < 0 || which >= MAXPRELOAD)
		myexit(1);
	if (!preload [which] || preload_len[which]<=0)
		myexit(2);

	sprintf (lengthstr, "Content-length: %d\r\n", preload_len[which]);
	write (fd, header, headerlen);
	write (fd, lengthstr, strlen(lengthstr));
	write (fd, "\r\n", 2);

	int len;
	len = write (fd, preload[which], preload_len[which]);
	close (fd);

	switch (which) {
	case PRELOAD_INDEX: 
		log_write ("   \\\n    -> Sent INDEX"); 
		break;
	case PRELOAD_FAVICON: 
		log_write ("   \\\n    -> Sent FAVICON"); 
		break;
	default:
		log_write ("   \\\n    -> Sent preload"); 
		break;
	}

	bool success = len == preload_len[which];

	if (success)
		log_writeln(NULL);
	else
		log_writeln(" (failed).");

	log_store ();
	return myexit(1);
}

int send_file (int fd, char *path)
{
	if (!path) 
		return false;

	int len;
	unsigned char *data = get_file (path, &len, 0);
	if (!data)
		return false;

	int len2;
	len2 = write (fd, data, len);
	close (fd);

	bool success = len == len2;

	log_write ("   \\\n    -> Sent file '"); 
	log_write (path);
	log_write ("'");
	if (success)
		log_writeln(NULL);
	else
		log_writeln(" (failed).");

	log_store ();
	return myexit(1);
}


// Send the results of running a script
// Optionally precede by some header stuff
//
int send_results (int fd, char *path, char *headerpath)
{
	if (!path) {
		if (headerpath)
			unlink (headerpath);
		return false;
	}
	/*----------*/

	int len = 0;
	unsigned char *data = get_file (path, &len, 0);
	unlink (path);
	if (!data)
		return false;

	int hdrlen;
	unsigned char *hdr=NULL;
	if (headerpath) {
		hdr = get_file (headerpath, &hdrlen, 0);
		if (headerpath)
			unlink (headerpath);
		if (!hdr)
			return false;
	}
	bool success = true;

	if (headerlen != write (fd, header, headerlen))
		success = false;
	else
	{
		if (headerpath) {
			if (hdrlen != write (fd, hdr, hdrlen))
				success = false;
		}

		if (success) {
			if (len != write (fd, data, len))
				success = false;
		}
	}
	close (fd);

	log_write ("   \\\n    -> Sent results (");
	log_write_num (len);
	log_write (") ");
	if (headerpath)
		log_write ("with extra header");
	if (success)
		log_writeln(NULL);
	else
		log_writeln(" (failed).");

	log_store ();
	return true;
}


int process_request (int fd, struct sockaddr_in *from)
{
	static unsigned char req [REQLEN];
	static unsigned char ipstr [20];
	int i;

	if (!from || fd<0)
		myexit(98);
	/*----------*/

	IP addr = from->sin_addr.s_addr;
	unsigned short prefix = 0xff & addr;
	unsigned short second = 0xff & (addr >> 8);
	unsigned short third = 0xff & (addr >> 16);
	unsigned short fourth = 0xff & (addr >> 24);
	time_t t = time(NULL);

	sprintf (ipstr, "%d.%d.%d.%d", prefix, second, third, fourth);

#if 0
	struct hostent *he = gethostbyaddr (from, sizeof(from), AF_INET);
	char *remote_host = he ? he->h_name : "???";
#else
	char *remote_host = "unknown";
#endif

	log_writeln ("\n-------------------------------------------------------------------------------");
	log_write ("Request from ");
	log_write (ipstr);
	log_write (", on ");
	log_writeln (ctime(&t));

	i = read (fd, req, REQLEN);
	if (i >= 0)
		req[i]=0;

	log_write ("Bytes read: ");
	log_writeln_num (i);

	/* Are they on our shitlist? If so, send feces.
	 */
	if (ip_block_all[prefix] ||
	    ip_block_sub1[prefix] == second ||
	    ip_block_sub2[prefix] == second)
	{
		char *s = strchr (req,'\n');
		if (s) {
			*s = 0;
			log_write ("Request from blocked IP was: ");
			log_writeln (req);
		}
		send_feces (fd);
	}

	if (i >= MAXREQ) {
		log_writeln ("\tERROR: request too long");
		log_store ();
		myexit(EXITCODE_BUFFER_ATTACK);
	}
	else
	if (i < 0) {
		log_write ("\tERROR: read error\n");
		log_store ();
		perror ("read");
		myexit(EXITCODE_READ_ERROR);
	}
	else
	if (i < 4) {
		/* In my experience, it is possible that a browser
		 * can generate a zero-byte request when the user
		 * repeatedly reloads. It is therefore best to call
		 * this a "bad request" rather than an attack.
		 */
		log_write ("\tERROR: very short request\n");
		log_store ();
		myexit(EXITCODE_SHORT);
	}

	req[i]=0;
	if (!strncasecmp (req, "post ", 5)) 
	{
		req[6] = 0;
		log_write ("\tERROR: non-GET request ");
		log_writeln (req);
		http_error (fd, HTTP_ERR_BAD_REQUEST);
	}
	else
	if (!strncasecmp (req, "head", 4))
	{
		req[6] = 0;
		log_write ("\tERROR: non-GET request ");
		log_writeln (req);
		http_error (fd, HTTP_ERR_BAD_REQUEST);
	}
	else
	if (strncasecmp (req, "get ", 4))
	{
		req[6] = 0;
		log_write ("\tERROR: non-GET request ");
		log_writeln (req);
		exit (EXITCODE_BADREQ);
	}

	// Parse the parameters
	StringPair pairs[50];
	int referer = -1;
	int param_flags = 0;
	int connection = 0;
	int total_pairs = parse_http_header (req, pairs, 50, &referer,
		&param_flags, &connection);

	switch (connection) {
	case 1: log_writeln ("connection type is close"); break;
	case 2: log_writeln ("connection type is keep-alive"); break;
	case 3: log_writeln ("connection type is OTHER"); break;
	}

	// Log the parameters
	if (!total_pairs)
		log_write ("\tno params\n");
	else {
		for (i=0; i<total_pairs; i++) {
			log_write ("Param ");
			log_write_num (i);
			log_write (": ");
			log_write (pairs[i].name);
			log_write (" = ");
			log_writeln (pairs[i].contents);
		}
	}

	if (param_flags & HEADER_HAS_AUTHORIZATION) {
		log_write ("\tERROR: unnecessary Authorization included\n" );
		log_store ();
		myexit(EXITCODE_IGNORE);
	}
	if (!param_flags & HEADER_BAD_USERAGENT) {
		log_write ("\tERROR: bad or missing User-agent\n" );
		log_store ();
		myexit(EXITCODE_PARAM_ATTACK);
	}

	if (total_pairs >= 40) {
		log_write ("\tERROR: improbably large number of params\n" );
		log_store ();
		myexit(EXITCODE_PARAM_ATTACK);
	}
	else if (total_pairs < 3) {
		log_write ("\tERROR: too few params\n" );
		log_store ();
		myexit(0 & EXITCODE_FEW_PARAMS);
	}

	// Inspect parameters
	// Extract the GET target.
#define MAXTARGETLEN 4096
	char target [MAXTARGETLEN + 2];
	int ix=0;
	char *s = req + 4;
	char *s2 = target;
	int ch;
	while (ch=*s, ch && (ch==' ' || ch=='\t'))
		s++;
	while (ix < MAXTARGETLEN) {
		ch = *s++;
		ix++;
		if (ch=='\n' || !ch || ch==' ' || ch=='\t')
			break;
		*s2++ = ch;
	}
	*s2 = 0;

	// Print target
	log_write ("GET target = '");
	log_write (target);
	log_writeln ("'");

	// Isolate any parameters
	char *params = strchr (target, '?');
	if (params) {
		*params = 0;
		params++;
	}

	// Provide the index
	//
	if ('/' == target[0] && !target[1] && index) 
		send_preload (fd, PRELOAD_INDEX);
	if ('/' == target[0] && tolower(target[1])=='i' && index) {
		if (!strncasecmp (target+2, "ndex.htm", 8))
			send_preload (fd, PRELOAD_INDEX);
	}
	if (strstr (target,"/favicon.ico"))
		send_preload (fd, PRELOAD_FAVICON);

	// Check for bizarre "http://" request
	if (!strncasecmp ("http:/", target, 6)) {
		send_witty (fd);
	}
	if (!strncasecmp ("/http:/", target, 7)) {
		send_witty (fd);
	}

	// Determine the extension if any.
	char *extension=NULL;
	ix = strlen (target) - 1;
	ch=0;
	while (ch = target[ix], ix > 0 && ch != '/' && ch != '.') {
		ix--;
	}
	if (ch == '.') {
		extension = &target[ix+1];
	}

	char extncode = 0;
	if (extension && *extension)
		extncode = parse_extension(extension);

	char localpath[MAXPATH];
	memset (localpath, 0, MAXPATH);

	// Clear the cookies
	setenv ("HTTP_COOKIE", "", 1);

	setenv ("REQUEST_METHOD", "GET", 1);
	setenv ("QUERY_STRING", params?params:"noparams=nothing", 1);

	if (extncode == EXTN_NONE) {
		log_write ("\tERROR: no extension\n");
		http_error (fd, HTTP_ERR_NOT_FOUND);
	}

	strcat (localpath, "/var/lapis/");

	setenv ("GATEWAY_INTERFACE", "CGI/1.1", 1);
	setenv ("HTTP_ACCEPT", "text/html", 1);
	setenv ("SERVER_PROTOCOL", "HTTP/1.0", 1);
	setenv ("SERVER_SOFTWARE", "Lapis", 1);
	setenv ("REMOTE_ADDR", ipstr, 1);
	setenv ("REMOTE_HOST", remote_host, 1);
	setenv ("HTTP_COOKIES", "", 1);
	setenv ("HTTP_REFERER", referer!=-1 ? pairs[referer].contents :
		"unknown", 1);

	// For now, some bogus stuff..
	setenv ("SERVER_NAME", "a.b.com", 1);
	setenv ("REMOTE_PORT", "9999", 1);
	setenv ("REMOTE_IDENT", "???", 1);

	if (*target == '/')
		strncat (localpath, target+1, 4000);
	else
		strncat (localpath, target, 4000);

	//----------------------------------------------
	// Verify that the HTML or script file exists.
	//
	int len = stat_file (localpath);
	if (len < 0) {
		log_write ("\tERROR: file not found\n");
		http_error (fd, HTTP_ERR_NOT_FOUND);
	}

	// Once target file is known to exist, process it.
	//
	switch (extncode) 
	{
#ifdef SUPPORT_PERL
	case EXTN_PL:
		if (!process_script (fd, localpath, 1))
			http_error (fd, HTTP_ERR_INTERNAL);
		break;
#endif

#ifdef SUPPORT_PYTHON
	case EXTN_PY:
		if (!process_script (fd, localpath, 3))
			http_error (fd, HTTP_ERR_INTERNAL);
		break;
#endif

#ifdef SUPPORT_PHP
	case EXTN_PHP:
		if (!process_script (fd, localpath, 2))
			http_error (fd, HTTP_ERR_INTERNAL);
		break;
#endif

#ifdef SUPPORT_LAPIS
	case EXTN_LAPIS:
		if (!process_lapis (fd, localpath))
			http_error (fd, HTTP_ERR_INTERNAL);
		break;
#endif

	case EXTN_HTML:
	case EXTN_JPG:
	case EXTN_TIF:
	case EXTN_GIF:
	case EXTN_PNG:
	case EXTN_MPG:
	case EXTN_AVI:
	case EXTN_MP3:
	case EXTN_OGG:
	case EXTN_JS:
		if (!send_file (fd, localpath)) {
			log_write ("\tERROR: file send failed\n");
			http_error (fd, HTTP_ERR_NOT_FOUND);
		} else {
			log_store ();
			myexit(0);
		}
		return 0;

	default:
		http_error (fd, HTTP_ERR_BAD_EXTN);
	}
	return 0;
}

bool process_script (int fd, char *localpath, char type)
{
	char cmd [4096];
	char inputpath [30];
	char outputpath [30];

	maketempfile (inputpath);
	maketempfile (outputpath);

	FILE *f = fopen (inputpath, "w");
	fprintf (f, "foo=1\n");
	fclose (f);

	char *ws=NULL;
	switch (type) {
	case 1: ws = "perl"; break;
	case 2: ws = "php"; break;
	case 3: ws = "python"; break;
	default:
		return false;
	}
	if (type==2)
		sprintf (cmd, "%s -f %s %s > %s", ws, localpath, inputpath, outputpath);
	else
		sprintf (cmd, "%s %s < %s > %s", ws, localpath, inputpath, outputpath);
	system(cmd);

	unlink (inputpath);

	if (!send_results (fd, outputpath, NULL))  {
		unlink (outputpath);
		log_write ("\tERROR: script results send failed\n");
		http_error (fd, HTTP_ERR_INTERNAL);
	}
	
	unlink (outputpath);
	log_store ();
	return myexit(0);
}


